import warnings

import six

from swimlane.core.cache import check_cache
from swimlane.core.fields.valueslist import ValuesListField
from swimlane.core.resolver import AppResolver
from swimlane.core.resources.record import Record, record_factory
from swimlane.core.resources.report import Report
from swimlane.utils import random_string, one_of_keyword_only
from swimlane.utils.version import requires_swimlane_version


def _cast_to_bulk(field, value):
    """Method used to temporarily allow Bulk Modify for Value Lists,
    Not intended for public use Patch will be removed at next release"""
    if isinstance(field, ValuesListField):
        if field.multiselect:
            value = value or []
            children = []
            for child in value:
                field.validate_value(child)
                children.append(field.cast_to_swimlane(child))
            return children
        field.validate_value(value)
        return field.cast_to_swimlane(value)

    else:
        if field.multiselect:
            value = value or []
            children = []
            for child in value:
                children.append(field.cast_to_report(child))
            return children
        return field.cast_to_report(value)


class RecordAdapter(AppResolver):
    """Handles retrieval and creation of Swimlane Record resources"""

    @check_cache(Record)
    @one_of_keyword_only('id', 'tracking_id')
    def get(self, key, value):
        """Get a single record by id

        Supports resource cache

        .. versionchanged:: 2.17.0 
            Added option to retrieve record by tracking_id

        Keyword Args:
            id (str): Full record ID
            tracking_id (str): Record Tracking ID

        Returns:
            Record: Matching Record instance returned from API

        Raises:
            TypeError: No id argument provided
        """
        if key == 'id':
            response = self._swimlane.request('get', "app/{0}/record/{1}".format(self._app.id, value))
            return Record(self._app, response.json())
        if key == 'tracking_id':
            response = self._swimlane.request('get', "app/{0}/record/tracking/{1}".format(self._app.id, value))
            return Record(self._app, response.json())

    def search(self, *filters, **kwargs):
        """Shortcut to generate a new temporary search report using provided filters and return the resulting records

        Args:
            *filters (tuple): Zero or more filter tuples of (field_name, operator, field_value)

        Keyword Args:
            keywords (list(str)): List of strings of keywords to use in report search
            limit (int): Set maximum number of returned Records, defaults to `Report.default_limit`. Set to 0 to return
                all records

        Notes:
            Uses a temporary Report instance with a random name to facilitate search. Records are normally paginated,
            but are returned as a single list here, potentially causing performance issues with large searches.

            All provided filters are AND'ed together

            Filter operators are available as constants in `swimlane.core.search`

        Examples:

            ::

                # Return records matching all filters with default limit

                from swimlane.core import search

                records = app.records.search(
                    ('field_name', 'equals', 'field_value'),
                    ('other_field', search.NOT_EQ, 'value')
                )

            ::

                # Run keyword search with multiple keywords
                records = app.records.search(keywords=['example', 'test'])

            ::

                # Return all records from app
                records = app.records.search(limit=0)


        Returns:
            :class:`list` of :class:`~swimlane.core.resources.record.Record`: List of Record instances returned from the
                search results
        """
        report = self._app.reports.build(
            'search-' + random_string(8),
            keywords=kwargs.pop('keywords', []),
            limit=kwargs.pop('limit', Report.default_limit)
        )

        for filter_tuples in filters:
            report.filter(*filter_tuples)

        return list(report)

    def create(self, **fields):
        """Create and return a new record in associated app and return the newly created Record instance

        Args:
            **fields: Field names and values to be validated and sent to server with create request

        Notes:
            Keyword arguments should be field names with their respective python values

            Field values are validated before sending create request to server

        Examples:
            Create a new record on an app with simple field names

            ::

                record = app.records.create(
                    field_a='Some Value',
                    someOtherField=100,
                    ...
                )

            Create a new record on an app with complex field names

            ::

                record = app.records.create(**{
                    'Field 1': 'Field 1 Value',
                    'Field 2': 100,
                    ...
                })

        Returns:
            Record: Newly created Record instance with data as returned from API response

        Raises:
            swimlane.exceptions.UnknownField: If any fields are provided that are not available on target app
            swimlane.exceptions.ValidationError: If any field fails validation before creation
        """
        new_record = record_factory(self._app, fields)

        new_record.save()

        return new_record

    def create_batch(self, *args, **kwargs):
        """Old method name before bulk_create, bulk_modify, and bulk_delete were added

        Calls :meth:`bulk_create` with whatever arguments have been passed and emits a DeprecationWarning

        .. deprecated:: 2.17.0
           Use :meth:`bulk_create` instead
        """
        warnings.warn(
            '"app.records.create_batch(...)" has been deprecated and will be removed in next major release. '
            'Use "app.records.bulk_create(...)" instead.',
            DeprecationWarning
        )
        return self.bulk_create(*args, **kwargs)

    @requires_swimlane_version('2.15')
    def bulk_create(self, *records):
        """Create and validate multiple records in associated app

        Args:
            *records (dict): One or more dicts of new record field names and values

        Notes:
            Requires Swimlane 2.15+

            Validates like :meth:`create`, but only sends a single request to create all provided fields, and does not
            return the newly created records

            Any validation failures on any of the records will abort the batch creation, not creating any new records

            Does not return the newly created records

        Examples:
            Create 3 new records with single request

            ::

                app.records.bulk_create(
                    {'Field 1': 'value 1', ...},
                    {'Field 1': 'value 2', ...},
                    {'Field 1': 'value 3', ...}
                )

        Raises:
            swimlane.exceptions.UnknownField: If any field in any new record cannot be found
            swimlane.exceptions.ValidationError: If any field in any new record fails validation
            TypeError: If no dict of fields was provided, or any provided argument is not a dict
        """

        if not records:
            raise TypeError('Must provide at least one record')

        if any(not isinstance(r, dict) for r in records):
            raise TypeError('New records must be provided as dicts')

        # Create local records from factory for initial full validation
        new_records = []

        for record_data in records:
            record = record_factory(self._app, record_data)
            record.validate()

            new_records.append(record)

        self._swimlane.request(
            'post',
            'app/{}/record/batch'.format(self._app.id),
            json=[r._raw for r in new_records]
        )

    @requires_swimlane_version('2.17')
    def bulk_modify(self, *filters_or_records, **kwargs):
        """Shortcut to bulk modify records
        
        .. versionadded:: 2.17.0
        
        Args:
            *filters_or_records (tuple) or (Record): Either a list of Records, or a list of filters.

        Keyword Args:
            values (dict): Dictionary of one or more 'field_name': 'new_value' pairs to update

        Notes:
            Requires Swimlane 2.17+

        Examples:

            ::

                # Bulk update records by filter

                app.records.bulk_modify(
                    # Query filters
                    ('Field_1', 'equals', value1),
                    ('Field_2', 'equals', value2),
                    ...
                    # New values for records
                    values={
                        "Field_3": value3,
                        "Field_4": value4,
                        ...
                    }
                )

                # Bulk update records

                record1 = app.records.get(tracking_id='APP-1')
                record2 = app.records.get(tracking_id='APP-2')
                record3 = app.records.get(tracking_id='APP-3')

                app.records.bulk_modify(record1, record2, record3, values={"Field_Name": 'new value'})

        Returns:
            :class:`string`: Bulk Modify Job ID
        """
        values = kwargs.pop('values', None)

        if kwargs:
            raise ValueError('Unexpected arguments: {}'.format(kwargs))

        if not values:
            raise ValueError('Must provide "values" as keyword argument')

        if not isinstance(values, dict):
            raise ValueError("values parameter must be dict of {'field_name': 'update_value'} pairs")

        _type = validate_filters_or_records(filters_or_records)

        data_dict = {}
        record_stub = record_factory(self._app)

        # build record_id list
        if _type is Record:
            data_dict['recordIds'] = [record.id for record in filters_or_records]

        # build filters
        else:
            filters = []
            for filter_tuples in filters_or_records:
                field = record_stub.get_field(filter_tuples[0])
                filters.append({
                    "fieldId": field.id,
                    "filterType": filter_tuples[1],
                    "value": field.get_report(filter_tuples[2])
                })
            data_dict['filters'] = filters

        # build modifications
        modifications = []
        for field_name, update_value in values.items():
            mod_field = record_stub.get_field(field_name)
            if not mod_field.bulk_modify_support:
                raise ValueError("Field '{}' of Type '{}', is not supported for bulk modify".format(field_name, mod_field.__class__.__name__))
            modifications.append({
                "fieldId": {
                    "value": mod_field.id,
                    "type": "Id"
                },
                "value": _cast_to_bulk(mod_field, update_value),
                "type": "Create"
            })
        data_dict['modifications'] = modifications
        response = self._swimlane.request('put', "app/{0}/record/batch".format(self._app.id), json=data_dict)

        # Update records if instances were used to submit bulk modify request after request was successful
        if _type is Record:
            for record in filters_or_records:
                for field, val in six.iteritems(values):
                    record[field] = val

        return response.text

    @requires_swimlane_version('2.17')
    def bulk_delete(self, *filters_or_records):
        """Shortcut to bulk delete records
        
        .. versionadded:: 2.17.0
        
        Args:
            *filters_or_records (tuple) or (Record): Either a list of Records, or a list of filters.
            
        Notes:
            Requires Swimlane 2.17+
            
        Examples:
        
            ::
            
                # Bulk delete records by filter
                app.records.bulk_delete(
                    ('Field_1', 'equals', value1),
                    ('Field_2', 'equals', value2)
                )
                
                # Bulk delete by record instances
                record1 = app.records.get(tracking_id='APP-1')
                record2 = app.records.get(tracking_id='APP-2')
                record3 = app.records.get(tracking_id='APP-3')
                app.records.bulk_delete(record1, record2, record3)

        Returns:
            :class:`string`: Bulk Modify Job ID
        """

        _type = validate_filters_or_records(filters_or_records)
        data_dict = {}

        # build record_id list
        if _type is Record:
            record_ids = []
            for record in filters_or_records:
                record_ids.append(record.id)
            data_dict['recordIds'] = record_ids

        # build filters
        else:
            filters = []
            record_stub = record_factory(self._app)
            for filter_tuples in filters_or_records:
                field = record_stub.get_field(filter_tuples[0])
                filters.append({
                    "fieldId": field.id,
                    "filterType": filter_tuples[1],
                    "value": field.get_report(filter_tuples[2])
                })
            data_dict['filters'] = filters

        return self._swimlane.request('DELETE', "app/{0}/record/batch".format(self._app.id), json=data_dict).text


def validate_filters_or_records(filters_or_records):
    """Validation for filters_or_records variable from bulk_modify and bulk_delete"""
    # If filters_or_records is empty, fail
    if not filters_or_records:
        raise ValueError('Must provide at least one filter tuples or Records')
    # If filters_or_records is not list of Record or tuple, fail
    if not isinstance(filters_or_records[0], (Record, tuple)):
        raise ValueError('Cannot provide both filter tuples and Records')
    # If filters_or_records is not list of either Record or only tuple, fail
    _type = type(filters_or_records[0])
    for item in filters_or_records:
        if not isinstance(item, _type):
            raise ValueError("Expected filter tuple or Record, received {0}".format(item))

    return _type



"""Tests for basic fields without substantial logic"""
import numbers

import pytest
import six

from swimlane.core.fields import resolve_field_class, _FIELD_TYPE_MAP, Field, _build_field_type_map
from swimlane.core.fields.base import ReadOnly, FieldCursor
from swimlane.exceptions import ValidationError
from swimlane.utils import get_recursive_subclasses


def test_text_field(mock_record):
    """Test automatic type coercion during set"""

    assert mock_record['Severity'] == '7'
    mock_record['Severity'] = int(mock_record['Severity']) + 1
    assert mock_record['Severity'] == '8'


def test_numeric_field(mock_record):
    """Test type validation"""

    assert isinstance(mock_record['Numeric'], numbers.Number)

    with pytest.raises(ValidationError):
        mock_record['Numeric'] = 'Not a number'

    current_value = mock_record['Numeric']
    mock_record['Numeric'] += 1
    assert mock_record['Numeric'] == current_value + 1


def test_tracking_id_field(mock_record):
    """Test Tracking Id is readonly"""

    assert mock_record['Tracking Id'] == 'RA-7'

    with pytest.raises(ValidationError):
        mock_record['Tracking Id'] = 'Other Tracking value'


def test_default_repr(mock_record):
    assert repr(mock_record._fields['Numeric']) == '<NumberField: 1>'


def test_resolve_field_class():
    """Test looking up field by field $type"""
    # Check for error on known invalid type
    with pytest.raises(KeyError):
        resolve_field_class({'$type': 'Not a valid type'})

    # Check all fields with a type are resolvable by their types
    for cls in get_recursive_subclasses(Field):
        if cls.field_type:
            if isinstance(cls.field_type, six.string_types):
                types = [cls.field_type]
            else:
                types = cls.field_type

            for field_type in types:
                assert resolve_field_class({'$type': field_type}) is cls


def test_error_on_invalid_field_type():
    """Make sure an exception is raised when building type map if field_type is not a tuple or string"""
    class Base(object):
        field_type = None

    class Invalid(Base):
        field_type = 12345

    with pytest.raises(ValueError):
        _build_field_type_map(Base)


@pytest.mark.parametrize(
    'field_class',
    [cls for cls in _FIELD_TYPE_MAP.values() if not issubclass(cls, ReadOnly)]
)
def test_all_fields_empty_value(mock_record, field_class):
    """Test setting fields to empty value works for all field classes"""
    # Get any not readonly field instance of provided field_class
    # Does not guarantee full scope of all field subtypes function as expected
    for field in mock_record._fields.values():
        if isinstance(field, field_class) and not field.readonly:
            del mock_record[field.name]

            swimlane = field.get_swimlane()
            python = field.get_python()
            if getattr(field, 'is_multiselect', False) or isinstance(python, FieldCursor):
                # Multi select fields use cursors or ordered sets in most cases
                assert swimlane is None
                assert len(python) == 0

            else:
                assert swimlane is python is None



import os
import string
import sys

import mock
import pytest
from pkg_resources import DistributionNotFound

from swimlane.core.resolver import SwimlaneResolver
from swimlane.exceptions import InvalidSwimlaneBuildVersion
from swimlane.utils import (
    random_string,
    get_recursive_subclasses,
    import_submodules,
    one_of_keyword_only
)
from swimlane.utils.version import (
    compare_versions,
    requires_swimlane_version,
    get_package_version
)


def test_random_string():
    generated_string = random_string(8)
    assert len(generated_string) == 8
    assert set(generated_string).issubset(set(string.ascii_letters + string.digits))

    generated_string = random_string(20, string.ascii_uppercase)
    assert len(generated_string) == 20
    assert set(generated_string).issubset(set(string.ascii_uppercase))


def test_get_recursive_subclasses():
    class Base(object):
        pass

    class DirectChild(Base):
        pass

    class IndirectChild(DirectChild):
        pass

    assert sorted([DirectChild, IndirectChild], key=lambda cls: cls.__name__) == sorted(get_recursive_subclasses(Base), key=lambda cls: cls.__name__)
    assert [IndirectChild] == get_recursive_subclasses(DirectChild)


def test_import_submodules():
    fixture_path = os.path.join(
        os.path.dirname(__file__),
        'fixtures'
    )
    sys.path.insert(0, fixture_path)

    imported_modules = import_submodules('import_pkg')
    assert len(imported_modules) == 2

    sample_module = imported_modules['import_pkg.sample_module']
    assert sample_module.fixture_value == 'fixture_value'

    sys.path.pop(0)


@pytest.mark.parametrize('inputs,expected', [
    ('2', 0),
    ('1', 1),
    ('3', -1),
    ('2.13', 0),
    ('2.12', 1),
    ('2.14', -1),
    ('2.13.3', -1),
    ('2.13.2', 0),
    ('2.13.2-173414', 0),
    ('2.13.2-173415', -1)
])
def test_compare_versions_no_zerofill(inputs, expected):
    """Test compare versions when only comparing sections of version included in both provided versions"""
    assert compare_versions(inputs, '2.13.2-173414') == expected


@pytest.mark.parametrize('inputs,expected', [
    ('2', 1),
    ('1', 1),
    ('3', -1),
    ('2.13', 1),
    ('2.12', 1),
    ('2.14', -1),
    ('2.13.3', -1),
    ('2.13.2', 1),
    ('2.13.2-173414', 0),
    ('2.13.2-173415', -1)
])
def test_compare_versions_with_zerofill(inputs, expected):
    """Test compare versions when comparing sections of version with default 0 for missing sections"""
    target = '2.13.2-173414'
    assert compare_versions(inputs, target, True) == expected
    assert compare_versions(target, inputs, True) == -expected


def test_get_package_version():
    mock_dist = mock.MagicMock()
    mock_dist.version = '1.2.3'

    with mock.patch('swimlane.utils.version.get_distribution', return_value=mock_dist):
        version = get_package_version()
        assert tuple(version.split('.')[:2]) > ('0', '0', '0')

    with mock.patch('swimlane.utils.version.get_distribution', side_effect=DistributionNotFound):
        assert get_package_version() == '0.0.0.dev'


@pytest.mark.parametrize('kwargs,expected', [
    [
        {},
        TypeError
    ],
    [
        {'a': 'a', 'b': 'b'},
        TypeError
    ],
    [
        {'other': 'value'},
        TypeError
    ],
    [
        {'a': 'value_a'},
        'value_a'
    ],
    [
        {'b': 'value_b'},
        'value_b'
    ]
])
def test_one_of_keyword_only(kwargs, expected):
    """Test decorator requires exactly one keyword argument from provided list of options"""

    valid_options = ('a', 'b')

    @one_of_keyword_only(*valid_options)
    def func(key, value):
        return key in valid_options and value == expected

    if expected is TypeError:
        with pytest.raises(expected):
            func(**kwargs)

    else:
        assert func(**kwargs)


class TestRequiresSwimlaneVersion(object):
    """Tests for the requires_swimlane_version decorator"""

    def _get_resolver(self, min_version, max_version):
        """Factory used in tests"""
        class TestResolver(SwimlaneResolver):
            @requires_swimlane_version(min_version, max_version)
            def func(self, *args, **kwargs):
                return args, kwargs

        return TestResolver

    @pytest.mark.parametrize('min_version,max_version', [
        ('2.13', '2.12'),
        ('2', '1.2.3'),
        (None, None)
    ])
    def test_invalid_version_range(self, min_version, max_version):
        """Test that requires_swimlane_version raises ValueError when receiving conflicting version range"""
        with pytest.raises(ValueError):
            self._get_resolver(min_version, max_version)

    @pytest.mark.parametrize('min_version,max_version', [
        ('2.13', '2.13'),
        ('2.12', '2.13'),
        (None, '2.13'),
        ('2.13', None),
        ('2', '3.2.1'),
        ('2.3.4', '3')
    ])
    def test_valid_version_range(self, min_version, max_version):
        """Test that requires_swimlane_version does not raise exceptions when given a valid version range"""
        self._get_resolver(min_version, max_version)

    @pytest.mark.parametrize('swimlane_version,min_version,max_version', [
        ('2.13', '2.14', None),
        ('2.13.2', '2.14', None),
        ('2.13.2', '2.13.3', None),
        ('2.13', None, '2.12'),
        ('2.13.2', None, '2.13'),
        ('2.13.2', None, '2.13.1'),
        ('2.13.2', '2.13', '2.13.1'),
        ('2.13.2', '2.13.3', '2.13.4'),
    ])
    def test_call_invalid_version(self, swimlane_version, min_version, max_version):
        """Test that InvalidServerVersion is raised when calling a method with an out of range Swimlane version"""
        mock_swimlane = mock.MagicMock()
        mock_swimlane.build_version = swimlane_version

        resolver_class = self._get_resolver(min_version, max_version)

        resolver = resolver_class(mock_swimlane)

        with pytest.raises(InvalidSwimlaneBuildVersion):
            resolver.func()

    @pytest.mark.parametrize('swimlane_version,min_version,max_version', [
        ('2.13', '2.12', None),
        ('2.13', '2.13', None),
        ('2.13.2', '2.12', None),
        ('2.13', None, '2.13'),
        ('2.13.2', None, '2.13.3'),
        ('2.13.2', '2.13', '2.14'),
        ('2.13.2', '2.13.0', '2.14'),
        ('2.13.2', '2.13.0', '2.14.1'),
        ('2.13', '2.13', '2.14.1'),
    ])
    def test_call_valid_version(self, swimlane_version, min_version, max_version):
        """Test that InvalidServerVersion is raised when calling a method with an out of range Swimlane version"""
        mock_swimlane = mock.MagicMock()
        mock_swimlane.build_version = swimlane_version

        resolver_class = self._get_resolver(min_version, max_version)

        resolver = resolver_class(mock_swimlane)

        args = (1, 2, 3)
        kwargs = {'a': 'A', 'b': 'B'}

        assert resolver.func(*args, **kwargs) == (args, kwargs)



import math
import pendulum
import datetime
import pytz

import pytest

from swimlane.core.fields.datetime import DatetimeField
from swimlane.exceptions import ValidationError

UTC = pendulum.timezone('UTC')

datetime_now = datetime.datetime.now(pytz.timezone('MST'))
# Mongo drops portion of microsecond, field truncates automatically for consistency
datetime_now = datetime_now.replace(microsecond=int(math.floor(datetime_now.microsecond / 1000) * 1000))
pendulum_now = pendulum.instance(datetime_now)

pendulum_interval = pendulum.interval(minutes=5)


@pytest.mark.parametrize('field_name,dt,expected_raw', [
    ('Incident Closed', pendulum_now, DatetimeField.format_datetime(pendulum_now)),
    (
        'Date Field',
        pendulum_now,
        DatetimeField.format_datetime(pendulum.Pendulum(pendulum_now.year, pendulum_now.month, pendulum_now.day))
    ),
    ('Time Field', pendulum_now, DatetimeField.format_datetime(pendulum_now)),
    ('Incident Duration', pendulum_interval, pendulum_interval.in_seconds() * 1000)
])
def test_raw_serialization(mock_record, field_name, dt, expected_raw):
    """Test that datetime field values are appropriately serialized to raw"""
    mock_record[field_name] = dt
    field_id = mock_record.app.get_field_definition_by_name(field_name)['id']
    assert mock_record._raw['values'][field_id] == expected_raw


@pytest.mark.parametrize('valid_date_obj', [
    datetime_now,
    datetime_now.date(),
    pendulum_now,
    pendulum_now.date()
])
def test_date_set(mock_record, valid_date_obj):
    """Test that the date subtype can be set to a Date or Datetime"""
    mock_record['Date Field'] = valid_date_obj
    assert mock_record['Date Field'] == pendulum_now.date()


@pytest.mark.parametrize('invalid_date_obj', [
    0,
    '2017-05-15',
    '2017-05-15T00:00:00Z',
    datetime_now.time(),
    pendulum_now.time()
])
def test_date_set_invalid(mock_record, invalid_date_obj):
    """Test providing invalid data to date field"""
    with pytest.raises(ValidationError):
        mock_record['Date Field'] = invalid_date_obj


@pytest.mark.parametrize('valid_time_obj', [
    datetime_now,
    datetime_now.astimezone(pytz.UTC).time(),
    pendulum_now,
    UTC.convert(pendulum_now).time()
])
def test_time_set(mock_record, valid_time_obj):
    """Test that the time subtype can be set to a Time or Datetime"""
    mock_record['Time Field'] = valid_time_obj
    assert mock_record['Time Field'] == UTC.convert(pendulum_now).time()


@pytest.mark.parametrize('invalid_time_obj', [
    0,
    '2017-05-15',
    '2017-05-15T00:00:00Z',
    datetime_now.date(),
    pendulum_now.date()
])
def test_time_set_invalid(mock_record, invalid_time_obj):
    """Test providing invalid data to date field"""
    with pytest.raises(ValidationError):
        mock_record['Time Field'] = invalid_time_obj


def test_strip_trailing_microseconds(mock_record):
    """Test automatic removal of microseconds section during set to match values returned from API"""
    field = 'Incident Created'
    now_with_microsecond = pendulum.now().replace(microsecond=123456)
    mock_record[field] = now_with_microsecond

    assert mock_record[field] != now_with_microsecond
    assert mock_record[field].microsecond == 123000
    assert abs((mock_record[field] - now_with_microsecond).total_seconds()) < .001

    now_sans_microsecond = now_with_microsecond.replace(microsecond=0)
    mock_record[field] = now_sans_microsecond

    assert mock_record[field] == now_sans_microsecond



import pendulum

from swimlane.core.resources.base import APIResource
from swimlane.core.resources.usergroup import UserGroup
from .base import CursorField, FieldCursor, ReadOnly


class RevisionCursor(FieldCursor):
    """An iterable object that automatically lazy retrieves and caches history data for a record from API"""

    def __init__(self, *args, **kwargs):
        super(RevisionCursor, self).__init__(*args, **kwargs)
        self.__retrieved = False

    def _evaluate(self):
        """Lazily retrieves, caches, and returns the list of record _revisions"""
        if not self.__retrieved:
            self._elements = self._retrieve_revisions()
            self.__retrieved = True

        return super(RevisionCursor, self)._evaluate()

    def _retrieve_revisions(self):
        """Retrieve and populate Revision instances from history API endpoint"""
        response = self._swimlane.request(
            'get',
            'history',
            params={
                'type': 'Records',
                'id': self._record.id
            }
        )

        raw_revisions = response.json()

        return [Revision(self._record, raw) for raw in raw_revisions]


class Revision(APIResource):
    """Encapsulates a single revision returned from a History lookup"""

    def __init__(self, record, raw):
        super(Revision, self).__init__(record._swimlane, raw)

        self.record = record

        self.modified_date = pendulum.parse(self._raw['modifiedDate'])
        self.revision_number = int(self._raw['revisionNumber'])
        self.status = self._raw['status']

        # UserGroupSelection, can't set as User without additional lookup
        self.user = UserGroup(self._swimlane, self._raw['userId'])

        # Avoid circular imports
        from swimlane.core.resources.record import Record
        self.version = Record(self.record.app, self._raw['version'])

    def __str__(self):
        return '{} ({})'.format(self.version, self.revision_number)


class HistoryField(ReadOnly, CursorField):

    field_type = 'Core.Models.Fields.History.HistoryField, Core'
    cursor_class = RevisionCursor
    bulk_modify_support = False



"""Tests for the ResourceCache class"""

import copy

import mock
import pytest

from swimlane.core.cache import ResourcesCache, check_cache
from swimlane.core.resources.base import APIResource
from swimlane.core.resources.record import Record


def test_len(mock_app, mock_record):
    """Test cache length includes all cached resources"""

    cache = ResourcesCache(10)

    assert len(cache) == 0

    cache.cache(mock_app)
    assert len(cache) == 1

    # Ignore duplicates
    cache.cache(mock_app)
    assert len(cache) == 1

    cache.cache(mock_record)
    assert len(cache) == 2


def test_separate_resource_caches(mock_app, mock_record):
    """Test each resource class has a separate queue of max size instead of a single global queue"""

    # Max cache size of 1 per resource type
    cache = ResourcesCache(1)

    cache.cache(mock_app)
    cache.cache(mock_record)

    assert len(cache) == 2

    # Create another record instance
    other_record = copy.copy(mock_record)
    other_record.id = mock_record.id + '123'

    # Check that only one record is still in the cache
    cache.cache(other_record)
    assert len(cache) == 2


def test_item_in_cache(mock_record):
    """Test checking if item exists in cache"""

    cache = ResourcesCache(5)

    assert mock_record not in cache

    cache.cache(mock_record)

    assert mock_record in cache


def test_get_item_from_cache(mock_record):
    """Test retrieving item from cache, and that item is a copy instead of reference to same instance"""

    cache = ResourcesCache(5)

    cache_key = (type(mock_record), 'id', mock_record.id)

    # Attempt to get before record is in cache
    with pytest.raises(KeyError):
        cached_record = cache[cache_key]

    cache.cache(mock_record)

    cached_record = cache[cache_key]

    assert cached_record == mock_record
    assert cached_record is not mock_record

    # Prove failure after clearing cache
    cache.clear()
    with pytest.raises(KeyError):
        cached_record = cache[cache_key]


def test_clear(mock_app, mock_record):
    """Test clearing individual and all resources from cache"""

    cache = ResourcesCache(5)

    cache.cache(mock_app)
    cache.cache(mock_record)

    assert len(cache) == 2

    # Clear by resource class type
    cache.clear(type(mock_app))

    assert len(cache) == 1
    assert mock_app not in cache
    assert mock_record in cache

    cache.cache(mock_app)

    assert len(cache) == 2

    # Clear all caches
    cache.clear()

    assert len(cache) == 0
    assert mock_app not in cache
    assert mock_record not in cache


def test_check_cache_decorator(mock_swimlane, mock_record):
    """Check that decorator prevents actual function call on cache hits and defers to normal call on cache misses"""

    expected_miss_sentinel = object()

    mock_func = mock.MagicMock()
    mock_func.return_value=expected_miss_sentinel
    mock_func.__name__ = 'mock_func'
    decorated_func = check_cache(Record)(mock_func)

    mock_adapter = mock.MagicMock()
    mock_swimlane.resources_cache = ResourcesCache(5)
    mock_adapter._swimlane = mock_swimlane

    # Record not in cache yet, should call the actual mock_func with original inputs
    assert decorated_func(mock_adapter, id=mock_record.id) is expected_miss_sentinel
    mock_func.assert_called_once_with(mock_adapter, id=mock_record.id)

    # Record is returned from cache, shouldn't call the actual mock_func again
    mock_swimlane.resources_cache.cache(mock_record)
    assert decorated_func(mock_adapter, id=mock_record.id) == mock_record
    assert mock_func.call_count == 1


def test_unsupported_api_resource_instance(mock_swimlane):
    """Test that APIResource instances not returning all cache details are ignored but don't fail"""
    cache = ResourcesCache(5)

    cache.cache(APIResource(mock_swimlane, {}))

    assert len(cache) == 0


@pytest.mark.parametrize('item', [
    object(),
    None,
    {},
    [],
    'string',
    123
])
def test_cache_unsupported_type(item):
    """Test attempting to cache a non-APIResource instance fails"""
    cache = ResourcesCache(5)

    with pytest.raises(TypeError):
        cache.cache(item)


@pytest.mark.parametrize('key', [
    object(),
    None,
    {},
    [],
    (),
    ('string', 'wrong_leading_type', 'value'),
    (object, 'wrong_leading_type', 'value'),
    (Record, 'wrong_length'),
    (Record, 'wrong_length', 'wrong_length', 'wrong_length'),
])
def test_cache_invalid_index_key(key):
    """Test exception raised when attempting lookup with invalid index key"""
    cache = ResourcesCache(5)

    with pytest.raises(TypeError):
        item = cache[key]



import pytest

from swimlane.core.resources.usergroup import User
from swimlane.core.resources.base import APIResource


def test_repr(mock_user):
    """Use User resource to test base __repr__"""
    assert repr(mock_user) == '<User: admin>'


@pytest.mark.parametrize('raw', [
    {},
    {'$type': 'Wrong Type'}
])
def test_type_checks(mock_swimlane, raw):
    """Use User class to test $type validation"""
    with pytest.raises(TypeError):
        User(mock_swimlane, raw)


def test_base_no_default_cache_index_keys(mock_swimlane):
    r = APIResource(mock_swimlane, {})

    with pytest.raises(NotImplementedError):
        r.get_cache_index_keys()




def test_values_list_single_select_field(mock_record):
    """Test a ValuesList field in single-select mode"""
    assert mock_record['Status'] == 'Open'

    # Attempt to set value to invalid option
    try:
        mock_record['Status'] = 'Not valid'
    except ValueError:
        assert mock_record['Status'] == 'Open'
    else:
        raise RuntimeError

    # Set to valid option
    mock_record['Status'] = 'Closed'
    assert mock_record['Status'] == 'Closed'


def test_values_list_multi_select_field(mock_record):
    """Test a ValuesList field in multi-select mode"""
    vl_cursor = mock_record['Values List']
    assert len(vl_cursor) == 2

    # Adding the same value multiple times is ignored
    vl_cursor.select('Option 3')
    assert len(vl_cursor) == 3
    vl_cursor.select('Option 3')
    assert len(vl_cursor) == 3

    # Get item by index
    assert vl_cursor[0] == 'Option 1'
    assert vl_cursor[2] == 'Option 3'

    # Remove element raises exception if not already added
    vl_cursor.deselect('Option 3')
    assert len(vl_cursor) == 2

    try:
        vl_cursor.deselect('Option 3')
    except KeyError:
        assert len(vl_cursor) == 2
    else:
        raise RuntimeError

    # Respects field's valid options and types, raising ValueError for invalid values
    try:
        vl_cursor.select('Not a valid option')
    except ValueError:
        assert len(vl_cursor) == 2
    else:
        raise RuntimeError

    # Field can be set directly to any iterable, overwriting current selection entirely
    # Also resets field to a fresh cursor on next access
    vl_original_values = list(mock_record['Values List'])
    mock_record['Values List'] = []
    assert len(mock_record['Values List']) == 0

    # All elements must pass validation, or entire set operation fails
    try:
        mock_record['Values List'] = ['Option 1', 'Not a valid option']
    except ValueError:
        pass
    else:
        raise RuntimeError

    assert len(mock_record['Values List']) == 0
    mock_record['Values List'] = vl_original_values
    assert len(mock_record['Values List']) == 2

    # Attempt to directly set to a non-iterable value
    try:
        mock_record['Values List'] = 'Option 1'
    except ValueError:
        pass
    else:
        raise RuntimeError


def test_cursor_repr(mock_record):
    assert repr(mock_record['Values List']) == '<MultiSelectCursor: RA-7["Values List"] (2)>'


def test_get_report(mock_record):
    """Test behavior of values list get_report()"""
    field = mock_record.get_field('Values List')

    assert field.get_report(field.get_python()) == [
        '58fae4c59173122945a7cff6',
        '58fae4eafef0eead26dee65c'
    ]



import pytest

from swimlane.core.resources.usergroup import UserGroup
from swimlane.exceptions import ValidationError


def test_usergroup_single_field(mock_record, mock_group, mock_user):
    """Test single-select UserGroupField"""

    # Get usergroup from UserGroupField
    usergroup = mock_record['Incident Owner']

    assert isinstance(usergroup, UserGroup)
    assert usergroup.id == '58de1d1c07637a0264c0ca71'
    assert usergroup.name == 'Everyone'

    # UserGroup comparisons with specific User/Group instances
    assert usergroup == mock_group

    # Set User, Group, or UserGroup
    mock_record['Incident Owner'] = mock_user
    assert mock_record['Incident Owner'] == mock_user

    try:
        mock_record['Incident Owner'] = 'Everyone'
    except ValidationError:
        pass
    else:
        raise RuntimeError

    assert mock_record['Incident Owner'] == mock_user


def test_usergroup_multi_field(mock_record, mock_group, mock_user):
    """Test multi-select UserGroupField"""

    field_name = 'User/Groups'

    orig_value = mock_record[field_name]

    mock_record[field_name] = [mock_group, mock_user]

    assert list(mock_record[field_name]) == [mock_group, mock_user]

    mock_record[field_name] = orig_value


def test_usergroup_from_report(mock_record):
    """Test workaround for extraneous data returned from report without any selected users/groups on a multiselect field"""

    field_name = 'User/Groups'

    field = mock_record._fields[field_name]

    field.set_swimlane([{'$type': 'Core.Models.Utilities.UserGroupSelection, Core'}])

    assert list(mock_record[field_name]) == []



from io import BytesIO

import pendulum

from swimlane.core.resources.base import APIResource


class Attachment(APIResource):
    """Abstraction of an attachment from an AttachmentsField

    Attributes:
        file_id (str): Full file ID used in download request URL
        filename (str): Attachment filename
        upload_date (pendulum.Pendulum): Pendulum datetime when attachment was uploaded
    """

    _type = 'Core.Models.Record.Attachment, Core'

    def __init__(self, swimlane, raw):
        super(Attachment, self).__init__(swimlane, raw)

        self.file_id = self._raw['fileId']
        self.filename = self._raw['filename']
        self.upload_date = pendulum.parse(self._raw['uploadDate'])

    def __str__(self):
        return str(self.filename)

    def __hash__(self):
        return hash(self.file_id)

    def download(self, chunk_size=1024):
        """Download attachment

        Args:
            chunk_size (int): Byte-size of chunked download request stream

        Returns:
            BytesIO: Stream ready for reading containing the attachment file contents
        """
        stream = BytesIO()

        response = self._swimlane.request(
            'get',
            'attachment/download/{}'.format(self.file_id),
            stream=True
        )

        for chunk in response.iter_content(chunk_size):
            stream.write(chunk)

        stream.seek(0)

        return stream



import itertools


class Cursor(object):

    def __init__(self):
        self._elements = []

    def __len__(self):
        return len(list(self._evaluate()))

    def __iter__(self):
        for element in self._evaluate():
            yield element

    def __getitem__(self, item):
        return self._evaluate()[item]

    def _evaluate(self):
        """Hook to allow lazy evaluation or retrieval of cursor's elements

        Defaults to simply returning list of self._elements
        """
        return self._elements


class PaginatedCursor(Cursor):
    """Handle paginated lists, exposes hooks to simplify retrieval and parsing of paginated data"""

    default_limit = 0
    page_size = 10

    def __init__(self, limit=None):
        super(PaginatedCursor, self).__init__()

        if limit is None:
            limit = self.default_limit

        self.__limit = limit

        if self.__limit:
            self.page_size = min(self.page_size, self.__limit)

    def _evaluate(self):
        """Lazily retrieve and paginate report results and build Record instances from returned data"""
        if self._elements:
            for element in self._elements:
                yield element
        else:
            for page in itertools.count():
                raw_elements = self._retrieve_raw_elements(page)

                for raw_element in raw_elements:
                    element = self._parse_raw_element(raw_element)
                    self._elements.append(element)
                    yield element
                    if self.__limit and len(self._elements) >= self.__limit:
                        break

                if any([
                    len(raw_elements) < self.page_size,
                    (self.__limit and len(self._elements) >= self.__limit)
                ]):
                    break

    def _retrieve_raw_elements(self, page):
        """Send request and return response for single page of data"""
        raise NotImplementedError

    def _parse_raw_element(self, raw_element):
        """Hook to override parsing individual raw elements just before yielding"""
        return raw_element



import glob
import os
import subprocess
import tempfile
import zipfile

print('Extracting packages')

zip_file = zipfile.ZipFile(os.path.dirname(__file__))
temp_dir = tempfile.mkdtemp()
zip_file.extractall(temp_dir)
os.chdir(temp_dir)

#print('Installing virtualenv')
#
#virtualenv_wheel = glob.glob(os.path.join('.', 'deps', 'virtualenv*'))[0]
#
#subprocess.check_call(['pip', 'install', '--no-index', '--no-deps', '--upgrade', '--force-reinstall', '--find-links=.', virtualenv_wheel])
#
#import virtualenv
#
#print('Creating shim env')
#
#venv_dir = os.path.join(os.path.expanduser('~'), '.swimlane-shim')
#virtualenv.create_environment(venv_dir)

print('Installing swimlane driver')

#venv_pip = os.path.join(venv_dir, 'Scripts', 'pip.exe')
paths = glob.glob(os.path.join('.', 'deps', '*'))
subprocess.check_call(['python', '-m', 'pip', 'install', '--no-index', '--no-deps', '--upgrade', '--force-reinstall', '--find-links=.'] + paths)

shim_code = '''
import os

def activate_shim_env(venv_dir):
    activate_file = os.path.join(venv_dir, 'Scripts', 'activate_this.py')
    with open(activate_file) as f:
        exec(f.read(), {'__file__': activate_file})
        
'''

'''
print('\n'.join([
    '==========',
    'To activate new swimlane driver shim, copy following code block to top of task scripts',
    '==========',
    shim_code,
    'activate_shim_env({!r})'.format(venv_dir),
    'import swimlane',
    '',
    '=========='
]))
'''



import mock
import pytest

from swimlane.core.resources.report import report_factory
from swimlane.exceptions import UnknownField


class TestReport(object):

    def test_repr(self, mock_report):
        assert repr(mock_report) == '<Report: search-12345678>'

    def test_filter(self, mock_report):
        """Test adding a filter to a report"""

        # Attempt to use an operand the reports don't understand
        with pytest.raises(ValueError):
            mock_report.filter('Tracking Id', 'Invalid Operand', 'RA-7')

        # Attempt to filter on an invalid field
        with pytest.raises(UnknownField):
            mock_report.filter('TrackingId', "equals", 'RA-7')

        assert len(mock_report._raw['filters']) == 0
        mock_report.filter('Tracking Id', 'equals', 'RA-7')

        assert len(mock_report._raw['filters']) == 1

    def test_parse_raw_element(self, mock_app, mock_record):
        """Test _build_record makes a retrieval to the full record"""
        with mock.patch.object(mock_app.records, 'get', return_value=mock_record):
            report = report_factory(mock_app, 'build_record')
            assert report._parse_raw_element(mock_record._raw) is mock_record

    def test_limit(self, mock_record, mock_app, mock_swimlane):
        with mock.patch.object(mock_swimlane, 'request') as mock_request:
            mock_response = mock.MagicMock()
            mock_request.return_value = mock_response

            mock_response.json.return_value = {
                '$type': 'API.Models.Search.GroupedSearchResults, API',
                'count': 2,
                'limit': 50,
                'offset': 0,
                'results': {
                    '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[Core.Models.Record.Record[], Core]], mscorlib',
                    '58e4bb4407637a0e4c4f9873': [mock_record._raw for i in range(50)]}}

            with mock.patch('swimlane.core.resources.report.Report._parse_raw_element', return_value=mock_record):
                report = report_factory(mock_app, 'limit_test', limit=5)
                assert len(report) == 5

                report = report_factory(mock_app, 'default_limit')
                assert len(report) == 50

                report = report_factory(mock_app, 'no_limit', limit=100)
                assert len(report) == 100

    def test_iteration(self, mock_report, mock_record, mock_swimlane):
        """Test iterating over report results"""
        with mock.patch.object(mock_swimlane, 'request') as mock_request:
            mock_response = mock.MagicMock()
            mock_request.return_value = mock_response

            mock_response.json.return_value = {
                '$type': 'API.Models.Search.GroupedSearchResults, API',
                'count': 1,
                'limit': 50,
                'offset': 0,
                'results': {
                    '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[Core.Models.Record.Record[], Core]], mscorlib',
                    '58e4bb4407637a0e4c4f9873': [mock_record._raw]}}

            assert mock_request.call_count == 0
            with mock.patch.object(mock_report, '_parse_raw_element', return_value=mock_record):

                results = list(mock_report)

                assert len(results) == 1
                assert results[0] == mock_record

                assert mock_request.call_count == 1

                # Run again to check cache records returned instead of making additional requests
                results = list(mock_report)
                assert len(results) == 1
                assert mock_request.call_count == 1





import mock
import pytest

from swimlane.core.resources.usergroup import Group, User


def test_group_list(mock_group, mock_swimlane):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = {'groups': [mock_group._raw for _ in range(3)]}

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        groups = mock_swimlane.groups.list()
        assert len(groups) == 3
        for group in groups:
            assert isinstance(group, Group)


def test_group_get(mock_group, mock_swimlane):
    mock_response = mock.MagicMock()

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        mock_response.json.return_value = mock_group._raw

        group = mock_swimlane.groups.get(id=mock_group.id)
        assert isinstance(group, Group)

        mock_response.json.return_value = [mock_group._raw]

        group = mock_swimlane.groups.get(name=mock_group.name)
        assert isinstance(group, Group)

        mock_response.json.return_value = []
        with pytest.raises(ValueError):
            mock_swimlane.groups.get(name=mock_group.name)


@pytest.mark.parametrize('kwargs', [
    {'unknown_arg': 'arg'},
    {'name': 'name', 'id': 'id'},
    {}
])
def test_group_get_invalid_args(mock_swimlane, kwargs):
    with pytest.raises(TypeError):
        mock_swimlane.groups.get(**kwargs)


def test_user_list(mock_user, mock_swimlane):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = {'users': [mock_user._raw for _ in range(3)]}

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        users = mock_swimlane.users.list()
        assert len(users) == 3
        for user in users:
            assert isinstance(user, User)


def test_user_get(mock_user, mock_swimlane):
    mock_response = mock.MagicMock()

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        # Single id match
        mock_response.json.return_value = mock_user._raw
        user = mock_swimlane.users.get(id=mock_user.id)
        assert user == mock_user

        # Single display name match
        mock_response.json.return_value = [mock_user._raw]
        user = mock_swimlane.users.get(display_name=mock_user.display_name)
        assert user == mock_user

        # No users matching display name
        mock_response.json.return_value = []
        with pytest.raises(ValueError):
            mock_swimlane.users.get(display_name=mock_user.display_name)

        # Too many users matching display name
        mock_response.json.return_value = [mock_user._raw, mock_user._raw]
        with pytest.raises(ValueError):
            mock_swimlane.users.get(display_name=mock_user.display_name)

        # Error decoding JSON response body
        mock_response.json.return_value = [mock_user._raw]
        mock_response.json.side_effect = ValueError
        with pytest.raises(ValueError):
            mock_swimlane.users.get(id=mock_user.id)


@pytest.mark.parametrize('kwargs', [
    {'unknown_arg': 'arg'},
    {'display_name': 'name', 'id': 'id'},
    {}
])
def test_user_get_invalid_args(mock_swimlane, kwargs):
    with pytest.raises(TypeError):
        mock_swimlane.users.get(**kwargs)



import pytest
from mock import patch
import mock
import copy

from swimlane.core.cache import ResourcesCache
from swimlane.core.resources.usergroup import UserGroup, User, GroupUsersCursor


def test_group_users(mock_group, mock_user, mock_swimlane):

    mock_user2 = copy.copy(mock_user)
    mock_user3 = copy.copy(mock_user)
    mock_user2.id = mock_user2.id+'123'
    mock_user3.id = mock_user3.id+'abc'
    mock_group.user_ids = [mock_user.id, mock_user2.id, mock_user3.id]

    with mock.patch.object(mock_swimlane.users, 'get') as mock_user_get:
        mock_user_get.side_effect = [mock_user, mock_user2, mock_user3]
        assert isinstance(mock_group.users, GroupUsersCursor)
        for user in mock_group.users:
            assert isinstance(user, User)
        assert list(mock_group.users) == [mock_user, mock_user2, mock_user3]


def test_get_usergroup_selection(mock_user, mock_group):
    assert mock_user.as_usergroup_selection() == {
        '$type': 'Core.Models.Utilities.UserGroupSelection, Core',
        'id': '58de1d1c07637a0264c0ca6a',
        'name': 'admin'
    }

    assert mock_group.as_usergroup_selection() == {
        '$type': 'Core.Models.Utilities.UserGroupSelection, Core',
        'id': '58de1d1c07637a0264c0ca71',
        'name': 'Everyone'
    }


def test_metamethods(mock_user, mock_group, mock_record):
    """Test handful of simple metamethods for id and hashing, sorting, uniqueness, etc."""
    # Equality by id
    assert mock_user != mock_group

    # Ordering by name (Everyone < admin)
    assert mock_group < mock_user

    # Ordering type support
    with pytest.raises(TypeError):
        mock_user < mock_record

    # Hash
    local_set = set()
    local_set.add(mock_user)

    assert mock_user in local_set

    local_set.add(mock_user)

    assert len(local_set) == 1

    # Str
    assert str(mock_user) == mock_user.name


def test_resolve(mock_swimlane, mock_user, mock_group):
    """Test resolving a generic UserGroup to more specific type"""
    ug_user = UserGroup(mock_user._swimlane, mock_user._raw)
    ug_group = UserGroup(mock_group._swimlane, mock_group._raw)

    # Test resolve is no-op when already resolved
    for ug in (mock_user, mock_group):
        assert ug.resolve() is ug

    with patch.object(mock_swimlane.users, 'get') as mock_user_get:
        with patch.object(mock_swimlane.groups, 'get') as mock_group_get:
            mock_group_get.return_value = mock_group
            mock_user_get.return_value = mock_user

            assert ug_user.resolve() is mock_user

            mock_user_get.side_effect = ValueError

            assert ug_group.resolve() is mock_group


def test_caching(mock_user, mock_group):
    """Test support for User/Group caching"""

    cache = ResourcesCache(5)

    cache.cache(mock_user)
    cache.cache(mock_group)

    assert len(cache) == 2
    assert mock_user in cache
    assert mock_group in cache



import pytest

from swimlane.exceptions import ValidationError


def test_min_range(mock_record):
    """Test minimum range validation"""
    mock_record['Numeric (Range)'] = 0

    with pytest.raises(ValidationError):
        mock_record['Numeric (Range)'] = -1

    assert mock_record['Numeric (Range)'] == 0


def test_max_range(mock_record):
    """Test maximum range validation"""
    mock_record['Numeric (Range)'] = 10

    with pytest.raises(ValidationError):
        mock_record['Numeric (Range)'] = 11

    assert mock_record['Numeric (Range)'] == 10



"""Tests for custom Swimlane errors"""
import mock
import pytest
from requests import HTTPError

from swimlane.core.client import SwimlaneAuth, Swimlane
from swimlane.exceptions import SwimlaneHTTP400Error, InvalidSwimlaneProductVersion


def test_request_handling(mock_swimlane):
    """Test error message and code for SwimlaneHTTP400Error class"""

    with mock.patch.object(mock_swimlane, '_session') as mock_session:
        mock_response = mock.MagicMock()

        # Test with non-400 error; Should return a standard HTTPError
        mock_response.status_code = 401
        mock_session.request.return_value.raise_for_status.side_effect = HTTPError(response=mock_response)

        try:
            mock_swimlane.request('get', '/somepage')
        except HTTPError as error:
            assert not isinstance(error, SwimlaneHTTP400Error)
        else:
            raise RuntimeError

        # Test various 400 response error codes and messages
        mock_response.status_code = 400
        mock_response.json.return_value = {
            'ErrorCode': -1,
            'Argument': None
        }

        try:
            mock_swimlane.request('get', '/somepage')
        except SwimlaneHTTP400Error as error:
            assert error.code == -1
            assert error.name == 'Unknown'
            assert error.argument is None
        else:
            raise RuntimeError

        mock_response.json.return_value = {
            'Argument': 'argument',
            'ErrorCode': 3002
        }

        try:
            mock_swimlane.request('get', 'somepage')
        except SwimlaneHTTP400Error as error:
            assert error.code == 3002
            assert error.name == 'RecordNotFound'
            assert error.argument == 'argument'
        else:
            raise RuntimeError

        # Test failure to parse response content
        mock_response.json.side_effect = ValueError

        try:
            mock_swimlane.request('get', 'somepage')
        except SwimlaneHTTP400Error as error:
            assert error.code == -1
            assert error.name == 'Unknown'
            assert error.argument is None
        else:
            raise RuntimeError


def test_lazy_settings():
    """Test accessing settings is evaluated lazily and cached after first retrieval"""
    with mock.patch.object(Swimlane, 'request') as mock_request:
        with mock.patch.object(SwimlaneAuth, 'authenticate', return_value=(None, {})):
            mock_response = mock.MagicMock()
            mock_request.return_value = mock_response

            # Only include apiVersion setting for current tests
            data = {
                'apiVersion': '2.15.0-1234'
            }
            mock_response.json.return_value = data

            mock_swimlane = Swimlane('http://host', 'user', 'pass', verify_server_version=False)

            assert mock_request.call_count == 0

            assert mock_swimlane.settings == data
            assert mock_swimlane.version == data['apiVersion']

            assert mock_request.call_count == 1


def test_server_version_checks():
    """Test that server version is checked by default, raising InvalidServerVersion exception when failing"""
    with mock.patch('swimlane.core.client.requests.Session', mock.MagicMock()):
        with mock.patch.object(SwimlaneAuth, 'authenticate', return_value=(None, {})):
            with mock.patch.object(Swimlane, 'settings', new_callable=mock.PropertyMock) as mock_settings:
                mock_settings.return_value = {'apiVersion': '2.18+4.0.0+123456'}

                with mock.patch('swimlane.core.client.compare_versions', return_value=0):
                    Swimlane('http://host', 'admin', 'password')

                with mock.patch('swimlane.core.client.compare_versions', return_value=1):
                    with pytest.raises(InvalidSwimlaneProductVersion):
                        Swimlane('http://host', 'admin', 'password')


def test_auth(mock_swimlane):
    """Test automatic auth request and header injection with automatic version handling"""
    with mock.patch.object(mock_swimlane._session, 'request') as mock_request:
        mock_response = mock.MagicMock()
        mock_request.return_value = mock_response

        # Swimlane v2.13-
        mock_response.json.return_value = {
            '$type': 'API.Models.Identity.AuthorizeModel, API',
            'active': False,
            'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'createdDate': '2017-03-31T09:10:52.717Z',
            'disabled': False,
            'displayName': 'admin',
            'groups': [{'$type': 'Core.Models.Base.Entity, Core',
                        'disabled': False,
                        'id': '58de1d1c07637a0264c0ca71',
                        'name': 'Everyone'}],
            'id': '58de1d1c07637a0264c0ca6a',
            'isAdmin': True,
            'isMe': True,
            'lastLogin': '2017-04-27T14:11:38.54Z',
            'lastPasswordChangedDate': '2017-03-31T09:10:52.536Z',
            'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'modifiedDate': '2017-03-31T09:10:52.76Z',
            'name': 'admin',
            'passwordComplexityScore': 3,
            'passwordResetRequired': False,
            'permission': {'$type': 'Core.Models.Security.PermissionMatrix, Core'},
            'roles': [{'$type': 'Core.Models.Base.Entity, Core',
                       'disabled': False,
                       'id': '58de1d1c07637a0264c0ca64',
                       'name': 'Administrator'}],
            'userName': 'admin',
            'users': []}

        mock_response.cookies.items.return_value = [('.AspNetCore.Identity.Application',
                                                     'CfDJ8Eq07zdLE8ZGtPyzBNIkJhIgJB4-EUThghuYsSpPGU6fiO0uXHSfU2LDOmrpxmfa8KAVjMI1JSL1YyXzCXavaTMT7ZBwUO9J6rJG3m2p3B8hHFMd4RGRQypqP-znY6VEJmrvVr6_ZSF8sx-E54FI2N5WQo7gfiqGVIX70WrqyTvbaU1spoGsTRsXs1BJaUeobhSrI8MWjouqXSDuhcTJjjGczq_LGlBkdzYpV1wzPJSiwWXs-ZN2pIJzfMOedAJRs0OzIvVbB8aUn0zKfUu5K3QHhKImdudDqsEadGtTLccygC6t9b6XqIueMGuOYnn7mmOUv6MJBwdzTCfh2Eg4ElYOB8pqZsedWSbXz1GYuTlTpLWihJTwADMKtucIX1myfM4M9DVng_P9yp6K2BHm9_2jEcGJhnJQ2zJBML8TpRpdwhzz3iKBhBohFPqudEJ535zaR4onSO7dFGwbfx-fyZ5E31BNGl70A--y3VcNHeDqzs13ylpR--4DykUTGRnoYjsDFuD4ZHopvNGA8aC0LszaYeVW0M0aUzKw8VbiJdB3xQr9u6wkwscdzG3DqmJGgA')]

        auth = SwimlaneAuth(mock_swimlane, 'admin', 'password')

        assert auth._login_headers == {
            'Cookie': '.AspNetCore.Identity.Application=CfDJ8Eq07zdLE8ZGtPyzBNIkJhIgJB4-EUThghuYsSpPGU6fiO0uXHSfU2LDOmrpxmfa8KAVjMI1JSL1YyXzCXavaTMT7ZBwUO9J6rJG3m2p3B8hHFMd4RGRQypqP-znY6VEJmrvVr6_ZSF8sx-E54FI2N5WQo7gfiqGVIX70WrqyTvbaU1spoGsTRsXs1BJaUeobhSrI8MWjouqXSDuhcTJjjGczq_LGlBkdzYpV1wzPJSiwWXs-ZN2pIJzfMOedAJRs0OzIvVbB8aUn0zKfUu5K3QHhKImdudDqsEadGtTLccygC6t9b6XqIueMGuOYnn7mmOUv6MJBwdzTCfh2Eg4ElYOB8pqZsedWSbXz1GYuTlTpLWihJTwADMKtucIX1myfM4M9DVng_P9yp6K2BHm9_2jEcGJhnJQ2zJBML8TpRpdwhzz3iKBhBohFPqudEJ535zaR4onSO7dFGwbfx-fyZ5E31BNGl70A--y3VcNHeDqzs13ylpR--4DykUTGRnoYjsDFuD4ZHopvNGA8aC0LszaYeVW0M0aUzKw8VbiJdB3xQr9u6wkwscdzG3DqmJGgA'}

        mock_inflight_request = mock.MagicMock()
        auth(mock_inflight_request)
        mock_inflight_request.headers.update.assert_called_once_with(auth._login_headers)

        # Swimlane v2.14+
        JWT_TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJuYW1laWQiOiJhTHFfcWlCWVJyVThoIiwidW5pcXVlX25hbWUiOiJhZG1pbiIsIm5iZiI6MTQ5MzMzMTQyNCwiZXhwIjoxNDkzMzM1MDI0LCJpYXQiOjE0OTMzMzE0MjQsImlzcyI6IlN3aW1sYW5lIiwiYXVkIjoiU3dpbWxhbmUifQ.w27D6JgYj6UuoTUivUmwNv8USqeieTTPwmmhJviiDRQ'

        mock_response.json.return_value = {
            '$type': 'API.Models.Identity.AuthorizeModel, API',
            'active': False,
            'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'createdDate': '2017-04-12T21:32:30.345Z',
            'disabled': False,
            'displayName': 'admin',
            'email': '',
            'favorites': {
                '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[System.String, mscorlib]], mscorlib]], mscorlib'},
            'groups': [{'$type': 'Core.Models.Base.Entity, Core',
                        'disabled': False,
                        'id': 'aLq_qiBYSzR1R',
                        'name': 'Everyone'}],
            'id': 'aLq_qiBYRrU8h',
            'isAdmin': True,
            'isMe': True,
            'lastLogin': '2017-04-27T16:17:04.5160226-06:00',
            'lastPasswordChangedDate': '2017-04-12T21:32:30.083Z',
            'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                               'id': 'aLq_qiBYRrU8h',
                               'name': 'admin'},
            'modifiedDate': '2017-04-27T22:17:04.5170289Z',
            'name': 'admin',
            'passwordResetRequired': False,
            'permission': {'$type': 'Core.Models.Security.PermissionMatrix, Core'},
            'roles': [{'$type': 'Core.Models.Base.Entity, Core',
                       'disabled': False,
                       'id': 'aLq_qiBYR2CvJ',
                       'name': 'Administrator'}],
            'token': JWT_TOKEN,
            'userName': 'admin',
            'users': []}

        auth = SwimlaneAuth(mock_swimlane, 'admin', 'password')

        assert auth._login_headers == {'Authorization': 'Bearer {}'.format(JWT_TOKEN)}

        mock_inflight_request = mock.MagicMock()
        auth(mock_inflight_request)
        mock_inflight_request.headers.update.assert_called_once_with(auth._login_headers)


def test_repr(mock_swimlane):
    assert repr(mock_swimlane) == '<Swimlane: admin @ http://host v3.0+5.0.0+123456>'


def test_cache_default_disabled(mock_swimlane, mock_record):
    """Test caching is disabled with default Swimlane arguments"""
    mock_swimlane.resources_cache.cache(mock_record)
    assert len(mock_swimlane.resources_cache) == 0


def test_old_version_breakdown():
    """Test that product version, build version, and build number produce expected values in old single value format"""
    with mock.patch('swimlane.core.client.requests.Session', mock.MagicMock()):
        with mock.patch.object(SwimlaneAuth, 'authenticate', return_value=(None, {})):
            with mock.patch.object(Swimlane, 'settings', new_callable=mock.PropertyMock) as mock_settings:
                version = '2.17.0-123456'
                mock_settings.return_value = {'apiVersion': version}

                sw = Swimlane('http://host', 'admin', 'password', verify_server_version=False)
                assert sw.version == version
                assert sw.product_version == '2.17.0'
                assert sw.build_version == '2.17.0'
                assert sw.build_number == '123456'


def test_new_version_breakdown():
    """Test that product version, build version, and build number produce expected values in new multi value format"""
    with mock.patch('swimlane.core.client.requests.Session', mock.MagicMock()):
        with mock.patch.object(SwimlaneAuth, 'authenticate', return_value=(None, {})):
            with mock.patch.object(Swimlane, 'settings', new_callable=mock.PropertyMock) as mock_settings:
                version = '2.18+4.0.0+123456'
                mock_settings.return_value = {'apiVersion': version}

                sw = Swimlane('http://host', 'admin', 'password', verify_server_version=False)
                assert sw.version == version
                assert sw.product_version == '2.18'
                assert sw.build_version == '4.0.0'
                assert sw.build_number == '123456'



from .base import Field, ReadOnly


class TrackingField(ReadOnly, Field):

    field_type = (
        'Core.Models.Fields.TrackingField, Core',
        'Core.Models.Fields.Tracking.TrackingField, Core'
    )
    bulk_modify_support = False



import numbers
import warnings

import mock
import pytest
import six

from swimlane.exceptions import UnknownField
from swimlane.core.adapters.record import validate_filters_or_records, _cast_to_bulk
from swimlane.core.fields.valueslist import ValuesListField
from swimlane.core.fields.datetime import DatetimeField

def test_get(mock_swimlane, mock_app, mock_record):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = mock_record._raw

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        assert mock_app.records.get(id='record_id') == mock_record


def test_get_by_tracking_id(mock_swimlane, mock_app, mock_record):
    mock_response = mock.MagicMock()
    # tracking_id = mock_record.tracking_id
    mock_response.json.return_value = mock_record._raw

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        assert mock_app.records.get(tracking_id='record_tracking_id') == mock_record


@pytest.mark.parametrize('kwargs', [
    {'unknown_arg': 'arg'},
    {}
])
def test_invalid_args(mock_app, kwargs):
    with pytest.raises(TypeError):
        mock_app.records.get(**kwargs)


def test_create(mock_swimlane, mock_app, mock_record):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = mock_record._raw

    primitives = six.string_types + (
         numbers.Number,
         list,
         tuple,
         set
    )

    fields = {}
    for field_name, field in six.iteritems(mock_record._fields):
        field_value = field.get_python()
        if isinstance(field_value, primitives) and not field.readonly:
            fields[field_name] = field_value

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        assert mock_app.records.create(**fields) == mock_record


@pytest.mark.parametrize('records,expected', [
    ([], TypeError),
    ([123], TypeError),
    ([{'Nonexistent': 'Value'}], UnknownField),
    ([{}], None),
    ([{'Status': 'Open'}], None),
    ([{'Status': 'Open'}, {'Status': 'Open'}], None)
])
def test_bulk_create(mock_app, records, expected):
    if expected is not None:
        with pytest.raises(expected):
            mock_app.records.bulk_create(*records)

    else:
        mock_app.records.bulk_create(*records)


def test_create_batch_deprecated(mock_app):
    """Verify DeprecationWarning is emitted for create_batch method"""
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter("always")
        mock_app.records.create_batch({}, {})

        assert len(caught_warnings) == 1
        assert issubclass(caught_warnings[-1].category, DeprecationWarning)
        assert "create_batch" in str(caught_warnings[-1].message)


def test_search(mock_swimlane, mock_app, mock_record):
    # Run first to cache swimlane user
    with mock.patch.object(mock_swimlane._session, 'request') as mock_request:
        mock_response = mock.MagicMock()
        mock_response.json.return_value = [{
            '$type': 'Core.Models.Identity.ApplicationUser, Core',
            'active': False,
            'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'createdDate': '2017-03-31T09:10:52.717Z',
            'disabled': False,
            'displayName': 'admin',
            'groups': [],
            'id': '58de1d1c07637a0264c0ca6a',
            'isAdmin': True,
            'isMe': False,
            'lastLogin': '2017-04-27T14:11:38.54Z',
            'lastPasswordChangedDate': '2017-03-31T09:10:52.536Z',
            'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'modifiedDate': '2017-03-31T09:10:52.76Z',
            'name': 'admin',
            'passwordComplexityScore': 3,
            'passwordHash': 'AQAAAAEAACcQAAAAEESp9LR0jN3qPF2fw5qWdyceYxbeBbawMW5AFt31dA5n3xX16MFJWsU/j82heenFww==',
            'passwordResetRequired': False,
            'roles': [],
            'userName': 'admin'}]

        mock_request.return_value = mock_response

        # Access property to ensure call
        user = mock_swimlane.user

    mock_response = mock.MagicMock()
    mock_response.json.return_value = {
        '$type': 'API.Models.Search.GroupedSearchResults, API',
        'count': 1,
        'limit': 50,
        'offset': 0,
        'results': {
            '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[Core.Models.Record.Record[], Core]], mscorlib',
            '58e4bb4407637a0e4c4f9873': [mock_record._raw]}}

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        with mock.patch('swimlane.core.adapters.report.Report._parse_raw_element', return_value=mock_record):
            assert mock_app.records.search(('Tracking Id', 'equals', 'RA-7')) == [mock_record]


def test_bulk_delete(mock_swimlane, mock_app, mock_record):
    # test that requests is called with proper object for filters
    with mock.patch.object(mock_swimlane, 'request') as mock_func:
        mock_app.records.bulk_delete(('Numeric', 'equals', 1))
    mock_func.assert_called_once_with('DELETE', "app/{0}/record/batch".format(mock_app.id), json=
    {
        'filters': [{
            'fieldId': 'aqkg3', 'filterType': 'equals', 'value': 1
        }],
    })

    # test that requests is called with proper object for records
    with mock.patch.object(mock_swimlane, 'request') as mock_func:
        mock_app.records.bulk_delete(mock_record)
    mock_func.assert_called_once_with('DELETE', "app/{0}/record/batch".format(mock_app.id), json=
    {
        'recordIds': ['58ebb22807637a02d4a14bd6']
    })

    # test value error
    with pytest.raises(ValueError):
        mock_app.records.bulk_delete(mock_record, ('Numeric', 'equals', 1))


def test_filters_or_records_validation(mock_record):
    # No values provided should raise ValueError
    with pytest.raises(ValueError):
        validate_filters_or_records([])

    # Invalid type should raise ValueError
    with pytest.raises(ValueError):
        validate_filters_or_records(["not supported type"])
    #
    with pytest.raises(ValueError):
        validate_filters_or_records([mock_record, ('Number', 'equals', 1)])


def test_bulk_modify_by_filter(mock_swimlane, mock_app):
    """Test bulk modify by filter tuples updates records"""
    # patch swimlane.requests method to assert_called_once_with
    with mock.patch.object(mock_swimlane, 'request') as mock_func:
        mock_app.records.bulk_modify(('Numeric', 'equals', 1), values={'Numeric': 2})
    mock_func.assert_called_once_with('put', "app/{0}/record/batch".format(mock_app.id), json=
        {'filters': [{
                'fieldId': 'aqkg3', 'filterType': 'equals', 'value': 1
            }],
            'modifications': [{
                    'fieldId': {
                        'type': 'Id',
                        'value': 'aqkg3'
                    },
                    'type': 'Create', 'value': 2
                }]
        }
    )


def test_bulk_modify_by_record(mock_swimlane, mock_app, mock_record):
    """Test bulk modify by record/list of records"""
    # patch swimlane.requests method to assert_called_once_with
    with mock.patch.object(mock_swimlane, 'request') as mock_func:
        mock_app.records.bulk_modify(mock_record, values={'Numeric': 2})
    mock_func.assert_called_once_with('put', "app/{0}/record/batch".format(mock_app.id), json=
        {'modifications': [{
            'fieldId': {
                'type': 'Id', 'value': 'aqkg3'
            },
            'type': 'Create', 'value': 2
        }],
            'recordIds': ['58ebb22807637a02d4a14bd6']
        }
    )
    # Ensure record field is being set
    assert mock_record['Numeric'] == 2


def test_bulk_modify_errors(mock_app, mock_record):
    """Test bulk modify for expected ValueError on invalid inputs"""
    # ValueError when passing in combination of filter tuples and records
    with pytest.raises(ValueError):
        mock_app.records.bulk_modify(mock_record, ('Numeric', 'equals', 1), values={'Numeric': 2})
    # ValueError when values kwarg is missing
    with pytest.raises(ValueError):
        mock_app.records.bulk_modify(mock_record, 2)
    # ValueError when values is not dict
    with pytest.raises(ValueError):
        mock_app.records.bulk_modify(mock_record, values=2)
    # ValueError when additional kwargs beyond values
    with pytest.raises(ValueError):
        mock_app.records.bulk_modify(mock_record, values={}, other_val={})
    # ValueError when using unsupported fields
    with pytest.raises(ValueError):
        mock_app.records.bulk_modify(mock_record, values={'PCAP Attachment': mock_record['PCAP Attachment']})


def test_bulk_format_patch(mock_record, mock_group, mock_user):

    # test if case multiselect (Values List)
    value_field = mock_record.get_field('Values List')
    value_cursor = mock_record['Values List']
    assert len(value_cursor) == 2
    assert _cast_to_bulk(value_field, value_cursor) != value_field.get_report(value_cursor)

    # test if case singleselect (Values List)

    mock_record['Status'] = 'Closed'
    single_value_field = mock_record.get_field('Status')
    assert _cast_to_bulk(single_value_field, 'Closed') != single_value_field.get_report('Closed')

    # test else case multiselect (User/Groups)

    group_value = mock_record['User/Groups']
    group_field = mock_record.get_field('User/Groups')
    mock_record['User/Groups'] = [mock_group, mock_user]
    assert len(group_value) == 2
    assert _cast_to_bulk(group_field, group_value) == group_field.get_report(group_value)

    # test else case singleselect

    single_group_field = mock_record._fields['User/Groups']
    single_group_field.set_swimlane([{'$type': 'Core.Models.Utilities.UserGroupSelection, Core'}])
    single_group_value = mock_record['User/Groups']
    assert _cast_to_bulk(single_group_field, single_group_value) == single_group_field.get_report(single_group_value)


from datetime import datetime

import mock

from swimlane.core.fields.history import RevisionCursor, Revision
from swimlane.core.resources.usergroup import UserGroup

raw_revision_data = [
    {'$type': 'Core.Models.History.Revision, Core',
     'modifiedDate': '2017-04-10T16:26:17.065Z',
     'revisionNumber': 3.0,
     'status': 'historical',
     'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                'id': '58de1d1c07637a0264c0ca6a',
                'name': 'admin'},
     'version': {'$type': 'Core.Models.Record.Record, Core',
                 'actionsExecuted': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e557c8b81ce3e2c6515fd1': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.389Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6ac5c88412c4acda3d66c': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.396Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6aca5293267c35188ea37': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.401Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6acc099097a98678ccfcf': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.407Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6bee81f119baf40a659b9': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.984Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c62f257f2c4994b84140': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.412Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c6536afac95743437254': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.419Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c83415177d0ad68ea961': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.426Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e7bb9ad6e4d34e8f557b2b': True,
                     '58e7bbba9bc82babdd655653': True,
                     '58e7bd86ad9ba1f93ef4afae': True,
                     '58e7bddc6ed1ceda2950b621': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.437Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}}},
                 'allowed': [],
                 'applicationId': '58e4bb4407637a0e4c4f9873',
                 'applicationRevision': 72.0,
                 'comments': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.Comments, Core]], mscorlib]], mscorlib'},
                 'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                   'id': '58de1d1c07637a0264c0ca6a',
                                   'name': 'admin'},
                 'createdDate': '2017-04-10T16:26:16.486Z',
                 'disabled': False,
                 'id': '58ebb22807637a02d4a14bd6',
                 'isNew': False,
                 'linkedData': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     'a99ut': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017},
                     'ac1oa': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 5,
                               'dow': 3,
                               'h': 23,
                               'm': 42,
                               'mm': 4,
                               'q': 2,
                               'w': 14,
                               'y': 2017},
                     'aiir8': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017}},
                 'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'},
                 'modifiedDate': '2017-04-10T16:26:16.999Z',
                 'referencedByIds': [],
                 'referencedRecordIds': [],
                 'sessionTimeSpent': 0,
                 'timeTrackingEnabled': True,
                 'totalTimeSpent': 0,
                 'trackingFull': 'RA-7',
                 'trackingId': 7.0,
                 'values': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e4bb4407637a0e4c4f9875': 'RA-7',
                     'a04rv': '10.0.1.184:56005:3172',
                     'a0g40': 'Alert Action',
                     'a0k7v': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bae8c61082e377cc4d82',
                               'value': 'Open'},
                     'a18qr': 'User.Activity.Failed Logins',
                     'a1r3v': 'Logon',
                     'a3rvw': '3172',
                     'a4l9p': 'logdec1',
                     'a5v09': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bafd83f418637cf91c3e',
                               'value': 'ESA Alert'},
                     'a5yg3': '3172',
                     'a6if8': '[{"lifetime": "60", "rid": "4531291", "payload": "7808", "size": "11956", "service": "0", "eth.src": "00:0C:29:5B:98:5E", "feed.name": "Swimlane_Feed", "medium": "1", "ip.dst": "178.238.235.143", "org.dst": "Contabo GmbH", "sessionid": "4534669", "latdec.dst": "51.2993", "eth.type": "2048", "ip.src": "10.0.1.183", "domain.dst": "contabo.host", "eth.dst": "00:50:56:A7:4D:79", "did": "packdec1", "longdec.dst": "9.491", "packets": "122", "streams": "2", "country.dst": "Germany", "time": 1491593904, "ip.proto": "1"}]',
                     'a7gm5': '<14>Apr  5 23:42:54 localhost CEF:0|RSA|Security Analytics ESA|10.4|Module_58e40ea9e4b070dfd2a59ad5_Alert|Alert Action|7|rt=2017-04-05T23:42Z id=890eb2b1-2d90-46a5-b1ed-6e58b408ed83 source=10.0.1.184:56005:3172  action=authentication failure;; client=sshd;; device_class=Unix;; device_ip=10.0.1.178;; device_type=crossbeamc;; did=logdec1;; ec_activity=Logon;; ec_outcome=Failure;; ec_subject=User;; ec_theme=Authentication;; esa_time=1491435774280;; event_cat_name=User.Activity.Failed Logins;; event_source_id=10.0.1.184:56005:3172;; header_id=0006;; host_src=178.238.235.143;; level=4;; medium=32;; msg_id=000103;; rid=3172;; sessionid=3172;; size=175;; time=1491435771;; user_dst=user;; username=0;;\x00',
                     'a7ira': 'User',
                     'a8vjm': 'Failure',
                     'a990o': '7',
                     'a99ut': '2017-04-10T16:26:17.457Z',
                     'aasnf': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bb144121bd73d8f53845',
                               'value': 'High'},
                     'ac1oa': '2017-04-05T23:42:54Z',
                     'acqdq': '1491435774280',
                     'af967': '890eb2b1-2d90-46a5-b1ed-6e58b408ed83',
                     'afbbv': '',
                     'afff3': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                               'id': '58de1d1c07637a0264c0ca6a',
                               'name': 'admin'},
                     'ahziu': '178.238.235.143',
                     'ai1zz': '4',
                     'aiir8': '2017-04-10T16:26:16.389Z',
                     'aio1w': 'sshd',
                     'aokp5': 'Unix',
                     'apf9i': '10.0.1.178',
                     'arnd6': 'authentication failure',
                     'arok4': 'crossbeamc',
                     'aw9gd': 'ip.src=10.0.1.178 || ip.dst=10.0.1.178',
                     'awmtq': 'Authentication',
                     'axdvr': '175'},
                 'visualizations': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.VisualizationData, Core]], mscorlib]], mscorlib'}}},
    {'$type': 'Core.Models.History.Revision, Core',
     'modifiedDate': '2017-04-10T16:26:16.998Z',
     'revisionNumber': 2.0,
     'status': 'historical',
     'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                'id': '58de1d1c07637a0264c0ca6a',
                'name': 'admin'},
     'version': {'$type': 'Core.Models.Record.Record, Core',
                 'actionsExecuted': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e557c8b81ce3e2c6515fd1': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.389Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6ac5c88412c4acda3d66c': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.396Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6aca5293267c35188ea37': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.401Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6acc099097a98678ccfcf': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.407Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c62f257f2c4994b84140': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.412Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c6536afac95743437254': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.419Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c83415177d0ad68ea961': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.426Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e7bb9ad6e4d34e8f557b2b': True,
                     '58e7bbba9bc82babdd655653': True,
                     '58e7bd86ad9ba1f93ef4afae': True,
                     '58e7bddc6ed1ceda2950b621': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.437Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}}},
                 'allowed': [],
                 'applicationId': '58e4bb4407637a0e4c4f9873',
                 'applicationRevision': 72.0,
                 'comments': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.Comments, Core]], mscorlib]], mscorlib'},
                 'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                   'id': '58de1d1c07637a0264c0ca6a',
                                   'name': 'admin'},
                 'createdDate': '2017-04-10T16:26:16.486Z',
                 'disabled': False,
                 'id': '58ebb22807637a02d4a14bd6',
                 'isNew': False,
                 'linkedData': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     'a99ut': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017},
                     'ac1oa': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 5,
                               'dow': 3,
                               'h': 23,
                               'm': 42,
                               'mm': 4,
                               'q': 2,
                               'w': 14,
                               'y': 2017},
                     'aiir8': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017}},
                 'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'},
                 'modifiedDate': '2017-04-10T16:26:16.954Z',
                 'referencedByIds': [],
                 'referencedRecordIds': [],
                 'sessionTimeSpent': 0,
                 'timeTrackingEnabled': True,
                 'totalTimeSpent': 0,
                 'trackingFull': 'RA-7',
                 'trackingId': 7.0,
                 'values': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e4bb4407637a0e4c4f9875': 'RA-7',
                     'a04rv': '10.0.1.184:56005:3172',
                     'a0g40': 'Alert Action',
                     'a0k7v': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bae8c61082e377cc4d82',
                               'value': 'Open'},
                     'a18qr': 'User.Activity.Failed Logins',
                     'a1r3v': 'Logon',
                     'a3rvw': '3172',
                     'a4l9p': 'logdec1',
                     'a5v09': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bafd83f418637cf91c3e',
                               'value': 'ESA Alert'},
                     'a5yg3': '3172',
                     'a7gm5': '<14>Apr  5 23:42:54 localhost CEF:0|RSA|Security Analytics ESA|10.4|Module_58e40ea9e4b070dfd2a59ad5_Alert|Alert Action|7|rt=2017-04-05T23:42Z id=890eb2b1-2d90-46a5-b1ed-6e58b408ed83 source=10.0.1.184:56005:3172  action=authentication failure;; client=sshd;; device_class=Unix;; device_ip=10.0.1.178;; device_type=crossbeamc;; did=logdec1;; ec_activity=Logon;; ec_outcome=Failure;; ec_subject=User;; ec_theme=Authentication;; esa_time=1491435774280;; event_cat_name=User.Activity.Failed Logins;; event_source_id=10.0.1.184:56005:3172;; header_id=0006;; host_src=178.238.235.143;; level=4;; medium=32;; msg_id=000103;; rid=3172;; sessionid=3172;; size=175;; time=1491435771;; user_dst=user;; username=0;;\x00',
                     'a7ira': 'User',
                     'a8vjm': 'Failure',
                     'a990o': '7',
                     'a99ut': '2017-04-10T16:26:17.457Z',
                     'aasnf': {'$type': 'Core.Models.Record.ValueSelection, Core',
                               'id': '58e7bb144121bd73d8f53845',
                               'value': 'High'},
                     'ac1oa': '2017-04-05T23:42:54Z',
                     'acqdq': '1491435774280',
                     'af967': '890eb2b1-2d90-46a5-b1ed-6e58b408ed83',
                     'afbbv': '',
                     'afff3': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                               'id': '58de1d1c07637a0264c0ca6a',
                               'name': 'admin'},
                     'ahziu': '178.238.235.143',
                     'ai1zz': '4',
                     'aiir8': '2017-04-10T16:26:16.389Z',
                     'aio1w': 'sshd',
                     'aokp5': 'Unix',
                     'apf9i': '10.0.1.178',
                     'arnd6': 'authentication failure',
                     'arok4': 'crossbeamc',
                     'aw9gd': 'ip.src=10.0.1.178 || ip.dst=10.0.1.178',
                     'awmtq': 'Authentication',
                     'axdvr': '175'},
                 'visualizations': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.VisualizationData, Core]], mscorlib]], mscorlib'}}},
    {'$type': 'Core.Models.History.Revision, Core',
     'modifiedDate': '2017-04-10T16:26:16.953Z',
     'revisionNumber': 1.0,
     'status': 'historical',
     'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                'id': '58de1d1c07637a0264c0ca6a',
                'name': 'admin'},
     'version': {'$type': 'Core.Models.Record.Record, Core',
                 'actionsExecuted': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e557c8b81ce3e2c6515fd1': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.389Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6ac5c88412c4acda3d66c': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.396Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6aca5293267c35188ea37': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.401Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6acc099097a98678ccfcf': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.407Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c62f257f2c4994b84140': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.412Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c6536afac95743437254': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.419Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e6c83415177d0ad68ea961': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.426Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}},
                     '58e7bddc6ed1ceda2950b621': {
                         '$type': 'Core.Models.Workflow.History.ActionHistory, Core',
                         'dateExecuted': '2017-04-10T16:26:16.437Z',
                         'userId': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'}}},
                 'allowed': [],
                 'applicationId': '58e4bb4407637a0e4c4f9873',
                 'applicationRevision': 72.0,
                 'comments': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.Comments, Core]], mscorlib]], mscorlib'},
                 'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                   'id': '58de1d1c07637a0264c0ca6a',
                                   'name': 'admin'},
                 'createdDate': '2017-04-10T16:26:16.486Z',
                 'disabled': False,
                 'id': '58ebb22807637a02d4a14bd6',
                 'isNew': False,
                 'linkedData': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     'a99ut': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017},
                     'aiir8': {'$type': 'Core.Models.Record.DateTimeParsed, Core',
                               'd': 10,
                               'dow': 1,
                               'h': 16,
                               'm': 26,
                               'mm': 4,
                               'q': 2,
                               'w': 15,
                               'y': 2017}},
                 'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                                    'id': '58de1d1c07637a0264c0ca6a',
                                    'name': 'admin'},
                 'modifiedDate': '2017-04-10T16:26:16.486Z',
                 'referencedByIds': [],
                 'referencedRecordIds': [],
                 'sessionTimeSpent': 0,
                 'timeTrackingEnabled': True,
                 'totalTimeSpent': 0,
                 'trackingFull': 'RA-7',
                 'trackingId': 7.0,
                 'values': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib',
                     '58e4bb4407637a0e4c4f9875': 'RA-7',
                     'a04rv': '10.0.1.184:56005:3172',
                     'a0g40': 'Alert Action',
                     'a18qr': 'User.Activity.Failed Logins',
                     'a1r3v': 'Logon',
                     'a3rvw': '3172',
                     'a4l9p': 'logdec1',
                     'a5yg3': '3172',
                     'a7gm5': '<14>Apr  5 23:42:54 localhost CEF:0|RSA|Security Analytics ESA|10.4|Module_58e40ea9e4b070dfd2a59ad5_Alert|Alert Action|7|rt=2017-04-05T23:42Z id=890eb2b1-2d90-46a5-b1ed-6e58b408ed83 source=10.0.1.184:56005:3172  action=authentication failure;; client=sshd;; device_class=Unix;; device_ip=10.0.1.178;; device_type=crossbeamc;; did=logdec1;; ec_activity=Logon;; ec_outcome=Failure;; ec_subject=User;; ec_theme=Authentication;; esa_time=1491435774280;; event_cat_name=User.Activity.Failed Logins;; event_source_id=10.0.1.184:56005:3172;; header_id=0006;; host_src=178.238.235.143;; level=4;; medium=32;; msg_id=000103;; rid=3172;; sessionid=3172;; size=175;; time=1491435771;; user_dst=user;; username=0;;\x00',
                     'a7ira': 'User',
                     'a8vjm': 'Failure',
                     'a990o': '7',
                     'a99ut': '2017-04-10T16:26:17.457Z',
                     'acqdq': '1491435774280',
                     'af967': '890eb2b1-2d90-46a5-b1ed-6e58b408ed83',
                     'afbbv': '',
                     'afff3': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core',
                               'id': '58de1d1c07637a0264c0ca6a',
                               'name': 'admin'},
                     'ahziu': '178.238.235.143',
                     'ai1zz': '4',
                     'aiir8': '2017-04-10T16:26:16.389Z',
                     'aio1w': 'sshd',
                     'aokp5': 'Unix',
                     'apf9i': '10.0.1.178',
                     'arnd6': 'authentication failure',
                     'arok4': 'crossbeamc',
                     'aw9gd': 'ip.src=10.0.1.178 || ip.dst=10.0.1.178',
                     'awmtq': 'Authentication',
                     'axdvr': '175'},
                 'visualizations': {
                     '$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.VisualizationData, Core]], mscorlib]], mscorlib'}}}]


def test_history_field(mock_record, mock_swimlane):
    history = mock_record['History']
    assert isinstance(history, RevisionCursor)

    mock_response = mock.MagicMock()
    mock_response.json.return_value = raw_revision_data

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):

        # Get number of revisions
        num_revisions = len(history)
        assert num_revisions == 3

        # Iterate backwards over revisions
        for idx, revision in enumerate(history):
            assert isinstance(revision, Revision)
            assert str(revision) == 'RA-7 ({})'.format(revision.revision_number)
            assert isinstance(revision.modified_date, datetime)
            assert isinstance(revision.user, UserGroup)
            assert revision.version.id == mock_record.id
            assert num_revisions - revision.revision_number == idx




"""Swimlane Python API driver"""

from __future__ import absolute_import

import logging

from .core.client import Swimlane
from .utils.version import get_package_version

__all__ = [
    'Swimlane',
]

__version__ = get_package_version()

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())




import weakref


class SwimlaneResolver(object):
    """Provides automatic weakref resolution for Swimlane client to avoid circular references and memory leaks"""

    def __init__(self, swimlane):
        self.__ref_swimlane = weakref.ref(swimlane)

    @property
    def _swimlane(self):
        """Transparently resolve the swimlane weakref"""
        return self.__ref_swimlane()


class AppResolver(SwimlaneResolver):
    """Provides automatic weakref resolution for Swimlane client and App instance"""

    def __init__(self, app):
        super(AppResolver, self).__init__(app._swimlane)

        self.__ref_app = weakref.ref(app)

    @property
    def _app(self):
        return self.__ref_app()



import pytest

from swimlane.core.cursor import Cursor, PaginatedCursor


def test_cursor_iteration():
    cursor = Cursor()
    cursor._elements = range(5)

    for idx, cursor_idx in zip(range(5), cursor):
        assert idx == cursor_idx


def test_cursor_len():
    cursor = Cursor()
    cursor._elements = range(5)

    assert len(cursor) == 5


def test_paginated_cursor_hooks():
    cursor = PaginatedCursor()

    # Default raises exception
    with pytest.raises(NotImplementedError):
        cursor._retrieve_raw_elements(0)

    # Noop parse hook
    assert cursor._parse_raw_element(1) == 1



import mock

app_id = '123'
record_id = '456'
field_id = '789'


def test_add_references(mock_swimlane):
    """Test the helper add-references route"""

    app_id = '123'
    record_id = '456'
    field_id = '789'
    target_record_ids = [
        app_id,
        record_id,
        field_id
    ]

    with mock.patch.object(mock_swimlane, 'request') as mock_request:
        mock_swimlane.helpers.add_record_references(
            app_id,
            record_id,
            field_id,
            target_record_ids
        )

        mock_request.assert_called_once_with(
            'post',
            'app/{}/record/{}/add-references'.format(app_id, record_id),
            json={
                'fieldId': field_id,
                'targetRecordIds': target_record_ids
            }
        )


def test_add_comment(mock_swimlane):
    with mock.patch.object(mock_swimlane, 'request') as mock_request:
        mock_swimlane.helpers.add_comment(
            app_id,
            record_id,
            field_id,
            'message'
        )

        assert mock_request.call_count == 1

def test_check_bulk_job_status(mock_swimlane):

    job_id = 'as03235as'
    with mock.patch.object(mock_swimlane, 'request') as mock_request:
        mock_swimlane.helpers.check_bulk_job_status(job_id)
    mock_request.assert_called_once_with('get', 'logging/job/{0}'.format(job_id))


from datetime import datetime
from io import BytesIO

import mock
import pytest
from swimlane.core.fields.attachment import AttachmentCursor
from swimlane.core.resources.attachment import Attachment
from swimlane.exceptions import ValidationError


def test_attachment_field(mock_swimlane, mock_record):
    mock_response = mock.MagicMock()

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):
        # AttachmentsField
        # Cursor managing iteration and addition of attachments
        attachments = mock_record['PCAP Attachment']
        assert isinstance(attachments, AttachmentCursor)
        assert len(attachments) == 1

        file_contents = b'file contents'

        mock_response.iter_content.return_value.__iter__.return_value = [file_contents]

        for attachment in attachments:
            assert isinstance(attachment, Attachment)
            assert attachment.filename == str(attachment) == '5f09afe50064b2bd718e77818b565df1.pcap'
            assert attachment.file_id == '58ebb22907637a0b488b7b17'
            assert isinstance(attachment.upload_date, datetime)
            stream = attachment.download()  # Retrieve file bytes as BytesIO stream (file-like object)
            assert isinstance(stream, BytesIO)
            content = stream.read()
            assert content == file_contents

        # Upload new attachment
        # Attachment is uploaded immediately, but not associated with record (on server) until saved
        mock_response.json.return_value = [{'$type': 'Core.Models.Record.Attachment, Core',
         'fileId': '58ebb22907637a0b488b7b17',
         'filename': '5f09afe50064b2bd718e77818b565df1.pcap',
         'uploadDate': '2017-04-10T16:26:17.017Z'}]

        new_attachment = attachments.add('filename.txt', BytesIO(b'file contents in stream/handle object'))
        assert isinstance(new_attachment, Attachment)
        assert len(attachments) == 2

        # Test adding another attachment
        attachments.add('next-filename.txt', BytesIO(b'file contents in stream/handle object'))
        assert len(attachments) == 3


def test_attachment_remove(mock_record):
    field_name = 'PCAP Attachment'
    field = mock_record.get_field(field_name)
    attachments = mock_record[field_name]
    # ensure attachments are present, and record values has the expected field id for the attachments field
    assert len(attachments) > 0
    assert field.id in mock_record._raw['values']

    # ensure ValidationError on setting to value other than 'None'
    with pytest.raises(ValidationError):
        mock_record[field_name] = 'invalid'

    mock_record[field_name] = None

    # assert new cursor
    newattachments = mock_record[field_name]
    assert len(newattachments) == 0
    assert newattachments is not attachments


def test_set_attachments_validation(mock_swimlane, mock_record):
    """Verify type validation when setting an attachments field"""
    attachment = Attachment(mock_swimlane, {
        '$type': 'Core.Models.Record.Attachment, Core',
        'fileId': '1234',
        'filename': 'filename.txt',
        'uploadDate': '2017-04-10T16:26:17.017Z'
    })

    field_name = 'PCAP Attachment'
    mock_record[field_name] = [attachment]
    assert list(mock_record[field_name]) == [attachment]

    with pytest.raises(ValidationError):
        mock_record[field_name] = [1, 2, 3]




import copy

import mock
import pytest

from swimlane.core.resources.record import Record, record_factory
from swimlane.exceptions import UnknownField, ValidationError


class TestRecord(object):
    def test_repr(self, mock_record):
        assert repr(mock_record) == '<Record: RA-7>'
        assert repr(record_factory(mock_record.app)) == '<Record: RA - New>'

    def test_save(self, mock_swimlane, mock_record):
        """Test save endpoint called with correct args"""

        with mock.patch.object(mock_swimlane, 'request') as mock_request:
            with mock.patch.object(mock_record, 'validate') as mock_validate:

                mock_request.return_value.json.return_value = mock_record._raw

                mock_record['Numeric'] = 5
                mock_record.save()

                mock_validate.assert_called_once_with()
                mock_request.assert_called_once_with(
                    'put',
                    'app/{}/record'.format(mock_record.app.id),
                    json={'$type': 'Core.Models.Record.Record, Core', 'actionsExecuted': {'$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib'}, 'allowed': [], 'applicationId': '58e4bb4407637a0e4c4f9873', 'applicationRevision': 0.0, 'comments': {'$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.Comments, Core]], mscorlib]], mscorlib', 'a3uau': [{'$type': 'Core.Models.Record.Comments, Core', 'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca6a', 'name': 'admin'}, 'createdDate': '2017-04-19T18:40:25.529Z', 'message': 'Example comment'}]}, 'createdDate': '0001-01-01T00:00:00', 'disabled': False, 'id': '58ebb22807637a02d4a14bd6', 'isNew': False, 'linkedData': {'$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib'}, 'modifiedDate': '0001-01-01T00:00:00', 'referencedByIds': [], 'referencedRecordIds': [], 'sessionTimeSpent': 0, 'timeTrackingEnabled': True, 'totalTimeSpent': 0, 'trackingId': 7.0, 'values': {'$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Object, mscorlib]], mscorlib', '58e4bb4407637a0e4c4f9875': 'RA-7', 'a04rv': '10.0.1.184:56005:3172', 'a0g40': 'Alert Action', 'a0k7v': {'$type': 'Core.Models.Record.ValueSelection, Core', 'id': '58e7bae8c61082e377cc4d82', 'value': 'Open'}, 'a11mn': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca6a', 'name': 'admin'}, 'a18qr': 'User.Activity.Failed Logins', 'a1k2a': 'https://app.threatconnect.com/auth/threat/threat.xhtml?threat=2058283', 'a1r3v': 'Logon', 'a2gor': 'Transparent Tribe', 'a2ybt': 'Germany', 'a365t': '2017-04-05T13:00:00.000000Z', 'a3rvw': '3172', 'a3w8f': ['58e24e8607637a0b488849c0', '58e24e8607637a0b488849ca', '58e24dc507637a0b48884950'], 'a44wz': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca71', 'name': 'Everyone'}, 'a4l9p': 'logdec1', 'a5g30': 'C2', 'a5v09': {'$type': 'Core.Models.Record.ValueSelection, Core', 'id': '58e7bafd83f418637cf91c3e', 'value': 'ESA Alert'}, 'a5yg3': '3172', 'a6if8': '[{"lifetime": "60", "rid": "4531291", "payload": "7808", "size": "11956", "service": "0", "eth.src": "00:0C:29:5B:98:5E", "feed.name": "Swimlane_Feed", "medium": "1", "ip.dst": "178.238.235.143", "org.dst": "Contabo GmbH", "sessionid": "4534669", "latdec.dst": "51.2993", "eth.type": "2048", "ip.src": "10.0.1.183", "domain.dst": "contabo.host", "eth.dst": "00:50:56:A7:4D:79", "did": "packdec1", "longdec.dst": "9.491", "packets": "122", "streams": "2", "country.dst": "Germany", "time": 1491593904, "ip.proto": "1"}]', 'a7bgv': '2017-04-21T00:00:00.000000Z', 'a7gm5': '<14>Apr  5 23:42:54 localhost CEF:0|RSA|Security Analytics ESA|10.4|Module_58e40ea9e4b070dfd2a59ad5_Alert|Alert Action|7|rt=2017-04-05T23:42Z id=890eb2b1-2d90-46a5-b1ed-6e58b408ed83 source=10.0.1.184:56005:3172  action=authentication failure;; client=sshd;; device_class=Unix;; device_ip=10.0.1.178;; device_type=crossbeamc;; did=logdec1;; ec_activity=Logon;; ec_outcome=Failure;; ec_subject=User;; ec_theme=Authentication;; esa_time=1491435774280;; event_cat_name=User.Activity.Failed Logins;; event_source_id=10.0.1.184:56005:3172;; header_id=0006;; host_src=178.238.235.143;; level=4;; medium=32;; msg_id=000103;; rid=3172;; sessionid=3172;; size=175;; time=1491435771;; user_dst=user;; username=0;;\x00', 'a7ira': 'User', 'a8vjm': 'Failure', 'a990o': '7', 'aabxm': 'Threat', 'aag0z': '<table class="table table-bordered table-hover"><tr><th>lifetime</th><th>rid</th><th>payload</th><th>size</th><th>service</th><th>eth.src</th><th>country.dst</th><th>medium</th><th>ip.dst</th><th>org.dst</th><th>sessionid</th><th>latdec.dst</th><th>eth.type</th><th>ip.src</th><th>domain.dst</th><th>eth.dst</th><th>did</th><th>longdec.dst</th><th>packets</th><th>streams</th><th>feed.name</th><th>time</th><th>ip.proto</th></tr><tr><td>60</td><td>4531291</td><td>7808</td><td>11956</td><td>0</td><td>00:0C:29:5B:98:5E</td><td>Germany</td><td>1</td><td>178.238.235.143</td><td>Contabo GmbH</td><td>4534669</td><td>51.2993</td><td>2048</td><td>10.0.1.183</td><td>contabo.host</td><td>00:50:56:A7:4D:79</td><td>packdec1</td><td>9.491</td><td>122</td><td>2</td><td>Swimlane_Feed</td><td>1491593904</td><td>1</td></tr></table>', 'aasnf': {'$type': 'Core.Models.Record.ValueSelection, Core', 'id': '58e7bb144121bd73d8f53845', 'value': 'High'}, 'aazhn': 'MYTHICLEOPARD', 'ac1oa': '2017-04-02T23:42:00.000000Z', 'acl9o': 'Mar 29 08:59:50 splunk lxcfs[940]:   4: fd:   9: net_cls,net_prio\nMar 29 08:59:50 splunk kernel: [    0.000000] Hypervisor detected: VMware\nMar 29 08:59:50 splunk kernel: [    0.000000] NODE_DATA(0) allocated [mem 0x13fff8000-0x13fffcfff]\nMar 29 08:59:50 splunk kernel: [    0.000000] RCU: Adjusting geometry for rcu_fanout_leaf=64, nr_cpu_ids=4\nMar 29 08:59:50 splunk kernel: [    0.000000] console [tty0] enabled\nMar 29 08:59:50 splunk kernel: [    0.187011] PM: Registering ACPI NVS region [mem 0xbfeff000-0xbfefffff] (4096 bytes)\nMar 29 08:59:50 splunk kernel: [    0.289673] pci 0000:00:15.4: [15ad:07a0] type 01 class 0x060400\nMar 29 08:59:50 splunk kernel: [    0.306003] pci 0000:00:18.5: System wakeup disabled by ACPI\nMar 29 08:59:50 splunk kernel: [    0.317367] pci 0000:00:15.4: PCI bridge to [bus 07]\nMar 29 08:59:50 splunk kernel: [    0.324226] pci 0000:00:18.3: PCI bridge to [bus 1e]\nMar 29 08:59:50 splunk kernel: [    0.374053] pci 0000:00:17.6: bridge window [io  0x1000-0x0fff] to [bus 19] add_size 1000\nMar 29 08:59:50 splunk kernel: [    0.376241] pci 0000:00:17.5: BAR 13: failed to assign [io  size 0x1000]\nMar 29 08:59:50 splunk kernel: [    0.381821] pci 0000:00:15.3: PCI bridge to [bus 06]\nMar 29 08:59:50 splunk kernel: [    0.389813] pci 0000:00:18.2: PCI bridge to [bus 1d]\nMar 29 08:59:50 splunk kernel: [    0.391570] pci_bus 0000:13: resource 2 [mem 0xeb800000-0xeb8fffff 64bit pref]\nMar 29 08:59:50 splunk kernel: [    0.788759] pcie_pme 0000:00:15.0:pcie01: service driver pcie_pme loaded\nMar 29 08:59:50 splunk kernel: [    0.796112] pciehp 0000:00:15.6:pcie04: service driver pciehp loaded\nMar 29 08:59:50 splunk kernel: [    0.852911] ehci-pci: EHCI PCI platform driver\nMar 29 08:59:50 splunk kernel: [    2.111909] AVX version of gcm_enc/dec engaged.\nMar 29 08:59:50 splunk kernel: [    4.099053] md: raid4 personality registered for level 4\nMar 29 08:59:50 splunk irqbalance[1106]:    ...done.\nMar 29 11:12:59 splunk sshd[3125]: fatal: Unable to negotiate with 10.0.1.10 port 55450: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nMar 29 19:56:21 splunk sshd[9399]: fatal: Unable to negotiate with 10.0.1.10 port 50808: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nMar 30 05:17:01 splunk CRON[16176]: pam_unix(cron:session): session opened for user root by (uid=0)\nMar 30 13:16:22 splunk systemd[1]: apt-daily.timer: Adding 1h 12min 7.795190s random time.\nMar 30 18:22:08 splunk sshd[26877]: fatal: Unable to negotiate with 10.0.1.10 port 47092: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nMar 31 03:17:01 splunk CRON[1496]: pam_unix(cron:session): session closed for user root\nMar 31 11:17:01 splunk CRON[7743]: pam_unix(cron:session): session closed for user root\nMar 31 19:36:54 splunk systemd[1]: apt-daily.timer: Adding 3h 28min 17.687266s random time.\nMar 31 20:17:01 splunk CRON[14688]: (root) CMD (   cd / && run-parts --report /etc/cron.hourly)\nApr  1 05:43:20 splunk systemd[1]: snapd.refresh.timer: Adding 3h 41min 31.130550s random time.\nApr  1 13:17:01 splunk CRON[27787]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  1 23:08:55 splunk sshd[2991]: fatal: Unable to negotiate with 10.0.1.10 port 45758: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nApr  2 06:47:01 splunk CRON[9119]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  2 06:47:03 splunk CRON[9119]: pam_unix(cron:session): session closed for user root\nApr  2 15:26:13 splunk sshd[15787]: fatal: Unable to negotiate with 10.0.1.10 port 51872: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nApr  3 00:17:01 splunk CRON[22704]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  3 08:58:35 splunk sshd[29536]: fatal: Unable to negotiate with 10.0.1.10 port 44038: no matching key exchange method found. Their offer: diffie-hellman-group1-sha1 [preauth]\nApr  3 18:17:01 splunk CRON[4260]: pam_unix(cron:session): session closed for user root\nApr  4 10:17:01 splunk CRON[15857]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  4 15:01:51 splunk sshd[19167]: pam_unix(sshd:auth): check pass; user unknown\nApr  4 15:12:20 splunk sshd[19342]: Failed password for invalid user admin from 10.0.3.200 port 61969 ssh2\nApr  4 15:23:20 splunk sshd[19546]: PAM 1 more authentication failure; logname= uid=0 euid=0 tty=ssh ruser= rhost=10.0.3.200\nApr  4 15:45:18 splunk sshd[19814]: pam_unix(sshd:auth): check pass; user unknown\nApr  4 16:08:34 splunk sshd[20033]: Invalid user admin from 10.0.3.200\nApr  4 22:17:01 splunk CRON[24590]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  5 11:17:01 splunk CRON[1667]: pam_unix(cron:session): session opened for user root by (uid=0)\nApr  5 11:17:01 splunk CRON[1667]: pam_unix(cron:session): session closed for user root\nApr  5 13:21:41 splunk sshd[3382]: message repeated 2 times: [ Failed password for user from 10.0.3.200 port 61496 ssh2]\nApr  5 13:35:07 splunk systemd-logind[996]: New session 202 of user user.\nApr  5 17:16:14 splunk systemd[1]: snapd.refresh.timer: Adding 3h 48min 41.276237s random time.\nApr  6 09:17:01 splunk CRON[18136]: pam_unix(cron:session): session closed for user root\nApr  6 09:17:10 splunk systemd[1]: Started Cleanup of Temporary Directories.\nApr  7 00:17:01 splunk CRON[29019]: pam_unix(cron:session): session closed for user root\nApr  7 11:17:01 splunk CRON[4922]: (root) CMD (   cd / && run-parts --report /etc/cron.hourly)\nApr  7 16:45:09 splunk sshd[9179]: Connection closed by 10.0.1.10 port 44167 [preauth]\nApr  7 22:40:33 splunk sshd[13666]: fatal: Unable to negotiate with 10.0.1.10 port 60949: no matching host key type found. Their offer: ecdsa-sha2-nistp521 [preauth]\nApr  7 23:17:01 splunk CRON[14063]: (root) CMD (   cd / && run-parts --report /etc/cron.hourly)\nApr  8 05:42:16 splunk sshd[19026]: Did not receive identification string from 10.0.1.10\nApr  8 10:48:13 splunk systemd[1]: snapd.refresh.timer: Adding 4h 27min 21.903828s random time.\nApr  8 17:17:01 splunk CRON[27791]: pam_unix(cron:session): session closed for user root\nApr  8 23:32:33 splunk sshd[32530]: Connection closed by 10.0.1.10 port 48600 [preauth]\nApr  8 23:32:33 splunk sshd[32533]: fatal: Unable to negotiate with 10.0.1.10 port 48615: no matching host key type found. Their offer: ecdsa-sha2-nistp384 [preauth]\nApr  9 05:25:42 splunk sshd[4731]: fatal: Unable to negotiate with 10.0.1.10 port 46071: no matching host key type found. Their offer: ecdsa-sha2-nistp384 [preauth]\nApr  9 11:22:54 splunk sshd[9292]: Protocol major versions differ for 10.0.1.10: SSH-2.0-OpenSSH_7.2p2 Ubuntu-4ubuntu2.1 vs. SSH-1.5-Nmap-SSH1-Hostkey\nApr  9 17:18:35 splunk sshd[13798]: Protocol major versions differ for 10.0.1.10: SSH-2.0-OpenSSH_7.2p2 Ubuntu-4ubuntu2.1 vs. SSH-1.5-NmapNSE_1.0\nApr  9 23:17:01 splunk CRON[18223]: pam_unix(cron:session): session closed for user root\nApr 10 02:17:01 splunk CRON[20422]: pam_unix(cron:session): session closed for user root\nApr 10 04:17:01 splunk CRON[22016]: pam_unix(cron:session): session closed for user root\nApr 10 06:32:03 splunk sshd[23779]: fatal: Unable to negotiate with 10.0.1.10 port 51063: no matching host key type found. Their offer: ecdsa-sha2-nistp521 [preauth]\nApr 10 09:18:01 splunk systemd[1]: Starting Cleanup of Temporary Directories...\nApr 10 09:18:01 splunk systemd[1]: Started Cleanup of Temporary Directories.\n', 'acqdq': '1491435774280', 'aep2e': [{'$type': 'Core.Models.Record.Attachment, Core', 'fileId': '58ebb22907637a0b488b7b17', 'filename': '5f09afe50064b2bd718e77818b565df1.pcap', 'uploadDate': '2017-04-10T16:26:17.017Z'}], 'af967': '890eb2b1-2d90-46a5-b1ed-6e58b408ed83', 'afbbv': '', 'afff3': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca6a', 'name': 'admin'}, 'afrn3': 'ThreatType/Targeted, KillChain/C2, MaliciousConfidence/High, Actor/MYTHICLEOPARD', 'agn03': [{'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca6a', 'name': 'admin'}, {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58ebc11107637a02d4a23f14', 'name': 'wade wilson'}], 'agoat': [{'$type': 'Core.Models.Record.Attachment, Core', 'fileId': '58ebb22a07637a0b488b7db3', 'filename': '362a36190a149866cf0f54e0fd119338.log', 'uploadDate': '2017-04-10T16:26:18.548Z'}], 'ah7js': 'high', 'ahdtf': 'http://mail.vdjunky.org/, http://www.vdjunky.org/, http://vdjunky.org/, http://178.238.235.143:9001/, http://178.238.235.143/, http://m1343.contabo.host/', 'ahep6': 'TLP RED', 'ahziu': '178.238.235.143', 'aiir8': '2017-04-10T16:26:16.389000Z', 'aio1w': 'sshd', 'akxvh': 177360000, 'alnet': 'ip_address', 'amjd0': 'ns1.contabo.net,ns2.contabo.net', 'anet4': '10.0.1.183, 178.238.235.143', 'ao64i': '2017-04-21T21:00:10.942000Z', 'aokp5': 'Unix', 'apf9i': '10.0.1.178', 'aqc6k': 'Contabo GmbH', 'aqkg3': 5, 'arlrt': [{'$type': 'Core.Models.Record.ValueSelection, Core', 'id': '58fae4eafef0eead26dee65c', 'value': 'Option 2'}, {'$type': 'Core.Models.Record.ValueSelection, Core', 'id': '58fae4c59173122945a7cff6', 'value': 'Option 1'}], 'arnd6': 'authentication failure', 'arok4': 'crossbeamc', 'au7ic': 'vdjunky.org, a6f8b4b528d67fe5b985ad0a394e46f5c116bb80b7cb8ca9a094d92f4dc614c1, 9b98abb9a9fa714e05d43b08b76c0afa, 85429d5f2745d813e53b28d3d953d1cd, 3e91836b89b6d6249741dc8ee0d2895a, tcp.vdjunky.org, 8253dd13c126c44483bcbde23e61d53d', 'aw9gd': 'ip.src=10.0.1.178 || ip.dst=10.0.1.178', 'awmtq': 'Authentication', 'axdvr': '175', 'ayqk6': '178.238.235.143', 'azfys': 2058283, 'azygs': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core', 'id': '58de1d1c07637a0264c0ca71', 'name': 'Everyone'}}, 'visualizations': {'$type': 'System.Collections.Generic.Dictionary`2[[System.String, mscorlib],[System.Collections.Generic.List`1[[Core.Models.Record.VisualizationData, Core]], mscorlib]], mscorlib'}})

                # Test validation failure
                mock_validate.side_effect = ValidationError(mock_record, 'Test error')

                with pytest.raises(ValidationError):
                    mock_record.save()

                assert mock_validate.call_count == 2
                assert mock_request.call_count == 1

    def test_delete(self, mock_swimlane, mock_record):
        """Test record can be deleted when not new and that deleted records lose their ID but maintain their fields"""
        with mock.patch.object(mock_swimlane, 'request') as mock_request:
            with mock.patch.object(mock_record, 'validate') as mock_validate:

                mock_record['Numeric'] = 5
                rec_id = mock_record.id

                mock_record.delete()
                mock_request.assert_called_once_with(
                    'delete',
                    'app/{}/record/{}'.format(mock_record.app.id, rec_id)
                )

                assert mock_record.id is None
                assert mock_record['Numeric'] == 5

                with pytest.raises(ValueError):
                    mock_record.delete()

    def test_validate_required_fields(self, mock_record):
        """Test validate method checks for required fields"""
        assert mock_record.validate() is None

        field_name = 'Numeric'
        mock_record[field_name] = None
        mock_record._fields[field_name].required = True

        with pytest.raises(ValidationError):
            mock_record.validate()

    def test_ordering(self, mock_record, mock_app):
        record_copy = Record(mock_record.app, mock_record._raw)

        # Equality by id and app id
        assert record_copy == mock_record

        record_copy.id = '58ebb22807637a02d4a14bd7'

        assert record_copy != mock_record

        # Ordering by tracking id and app name
        record_copy.tracking_id = 'RA-16'
        assert mock_record < record_copy

        # Verify can't sort/order against non-record instances
        with pytest.raises(TypeError):
            mock_record < mock_app

    def test_incorrect_app_id_failure(self, mock_app):
        """Test that retrieving a record validates source app has expected application ID"""
        with pytest.raises(ValueError):
            Record(mock_app, {
                '$type': Record._type,
                'isNew': False,
                'applicationId': mock_app.id + '1234',
            })

    def test_unknown_field_access(self, mock_record):
        """Test accessing a missing field raises UnknownField"""

        # Get
        try:
            mock_record['Muneric']
        except UnknownField as error:
            assert error.field_name == 'Muneric'
            assert error.similar_field_names == ['Numeric']
            assert error.app is mock_record.app
        else:
            raise RuntimeError

        # Set
        try:
            mock_record['Muneric'] = 5
        except UnknownField as error:
            assert error.field_name == 'Muneric'
            assert error.similar_field_names == ['Numeric']
            assert error.app is mock_record.app
        else:
            raise RuntimeError

    def test_iteration(self, mock_record):
        """Test that iterating over a record yields field names and their values like dict.items()"""
        num_fields = 0

        for field_name, field_value in mock_record:
            num_fields += 1
            assert field_value == mock_record[field_name]

        assert num_fields > 0

    def test_add_restriction(self, mock_swimlane, random_mock_user, mock_record):
        """Test adding usergroup restrictions to a record"""

        with pytest.raises(TypeError):
            mock_record.add_restriction()

        with pytest.raises(TypeError):
            mock_record.add_restriction(object())

        assert len(mock_record._raw['allowed']) == 0
        assert len(mock_record.restrictions) == 0

        mock_record.add_restriction(mock_swimlane.user)
        assert mock_swimlane.user in mock_record.restrictions
        assert len(mock_record._raw['allowed']) == 1

        mock_record.add_restriction(mock_swimlane.user)
        assert len(mock_record.restrictions) == 1

        mock_record.add_restriction(random_mock_user)
        assert len(mock_record.restrictions) == 2


    def test_remove_restriction(self, mock_swimlane, random_mock_user, mock_record):
        """Test adding usergroup restrictions to a record"""

        assert len(mock_record.restrictions) == 0
        mock_record.add_restriction(mock_swimlane.user, random_mock_user)

        with pytest.raises(TypeError):
            mock_record.remove_restriction(object())

        mock_record.remove_restriction()
        assert len(mock_record.restrictions) == 0

        with pytest.raises(ValueError):
            mock_record.remove_restriction(mock_swimlane.user)

        mock_record.add_restriction(mock_swimlane.user, random_mock_user)

        mock_record.remove_restriction(mock_swimlane.user)

        assert random_mock_user in mock_record.restrictions

    def test_ignore_caching_when_unsaved(self, mock_record):
        """Test cache index keys are only returned when record has id, disabling caching when unsaved"""

        assert mock_record.get_cache_index_keys()

        mock_record.id = None
        with pytest.raises(NotImplementedError):
            mock_record.get_cache_index_keys()

    def test_record_field_resolvable_by_key(self, mock_record):
        """Test that fields can be resolved by their keys"""
        assert mock_record.get_field('Action') is mock_record.get_field('action-key')

    def test_record_keys_not_included_in_field_names(self, mock_record):
        """Make sure fields keys don't show up when iterating a record"""
        assert 'action-key' not in dict(mock_record)



from datetime import datetime

import mock

from swimlane.core.fields.comment import CommentCursor
from swimlane.core.resources.comment import Comment
from swimlane.core.resources.usergroup import UserGroup


def test_comment_field(mock_record, mock_swimlane):
    # CommentsField
    # Cursor managing iteration and addition of comments
    # Automatically added to record during save without manual call to create comment
    comments = mock_record['Investigation Summary']
    assert isinstance(comments, CommentCursor)
    assert len(comments) == 1

    for comment in comments:
        assert isinstance(comment, Comment)
        assert isinstance(comment.message, str)
        assert str(comment) == comment.message
        assert isinstance(comment.user, UserGroup)
        assert isinstance(comment.created_date, datetime)

    # Add new comment
    with mock.patch.object(mock_swimlane._session, 'request') as mock_request:
        mock_response = mock.MagicMock()
        mock_response.json.return_value = [{
            '$type': 'Core.Models.Identity.ApplicationUser, Core',
            'active': False,
            'createdByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'createdDate': '2017-03-31T09:10:52.717Z',
            'disabled': False,
            'displayName': 'admin',
            'groups': [],
            'id': '58de1d1c07637a0264c0ca6a',
            'isAdmin': True,
            'isMe': False,
            'lastLogin': '2017-04-27T14:11:38.54Z',
            'lastPasswordChangedDate': '2017-03-31T09:10:52.536Z',
            'modifiedByUser': {'$type': 'Core.Models.Utilities.UserGroupSelection, Core'},
            'modifiedDate': '2017-03-31T09:10:52.76Z',
            'name': 'admin',
            'passwordComplexityScore': 3,
            'passwordHash': 'AQAAAAEAACcQAAAAEESp9LR0jN3qPF2fw5qWdyceYxbeBbawMW5AFt31dA5n3xX16MFJWsU/j82heenFww==',
            'passwordResetRequired': False,
            'roles': [],
            'userName': 'admin'}]

        mock_request.return_value = mock_response

        comments.comment('New comment message')

        # Not persisted until saved, but still listed on local record
        assert len(comments) == 2
        assert comments[1].message == 'New comment message'



import weakref

from swimlane.core.resolver import SwimlaneResolver
from swimlane.exceptions import ValidationError


class Field(SwimlaneResolver):
    """Base class for abstracting Swimlane complex types"""

    field_type = None

    # Sentinel representing a field that has no current value
    _unset = object()

    # List of supported types, leave blank to disable type validation
    supported_types = []

    # Checks if bulk modify is supported for field
    bulk_modify_support = True

    def __init__(self, name, record):
        """Value not included during instantiation to prevent ambiguity between python and swimlane representations"""
        super(Field, self).__init__(record._swimlane)

        self.name = name
        self.__record_ref = weakref.ref(record)
        self._value = self._unset

        self.field_definition = self.record.app.get_field_definition_by_name(self.name)
        self.key = self.field_definition.get('key')
        self.id = self.field_definition['id']
        self.input_type = self.field_definition.get('inputType')
        self.required = self.field_definition.get('required', False)
        self.readonly = bool(self.field_definition.get('formula', self.field_definition.get('readOnly', False)))
        self.multiselect = self.field_definition.get('selectionType', 'single') == 'multi'

    def __repr__(self):
        return '<{class_name}: {py!r}>'.format(class_name=self.__class__.__name__, py=self.get_python())

    @property
    def record(self):
        """Resolve weak reference to parent record"""
        return self.__record_ref()

    def _get(self):
        """Default getter used for both representations unless overridden"""
        return self._value

    def get_python(self):
        """Return best python representation of field value"""
        return self._get()

    def get_swimlane(self):
        """Return best swimlane representation of field value"""
        return self.cast_to_swimlane(self._get())

    def get_report(self, value):
        """Return provided field Python value formatted for use in report filter"""
        if self.multiselect:
            value = value or []
            children = []

            for child in value:
                children.append(self.cast_to_report(child))

            return children

        return self.cast_to_report(value)

    def cast_to_python(self, value):
        """Called during set_swimlane, should accept a single raw value as provided from API

        Defaults to no-op
        """
        return value

    def cast_to_swimlane(self, value):
        """Called during get_swimlane, should accept a python value and return swimlane representation

        Defaults to no-op
        """
        return value

    def cast_to_report(self, value):
        """Cast single value to report format, defaults to cast_to_swimlane(value)"""
        return self.cast_to_swimlane(value)

    def validate_value(self, value):
        """Validate value is an acceptable type during set_python operation"""
        if self.readonly:
            raise ValidationError(self.record, "Cannot set readonly field '{}'".format(self.name))
        if value not in (None, self._unset):
            if self.supported_types and not isinstance(value, tuple(self.supported_types)):
                raise ValidationError(self.record, "Field '{}' expects one of {}, got '{}' instead".format(
                    self.name,
                    ', '.join([repr(t.__name__) for t in self.supported_types]),
                    type(value).__name__)
                )

    def _set(self, value):
        """Default setter used for both representations unless overridden"""
        self._value = value
        self.record._raw['values'][self.id] = self.get_swimlane()

    def set_python(self, value):
        """Set field internal value from the python representation of field value"""
        self.validate_value(value)
        return self._set(value)

    def set_swimlane(self, value):
        """Set field internal value from the swimlane representation of field value"""
        return self._set(self.cast_to_python(value))
